/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#pragma implementation


#ifdef QCL_USE_READLINE
#include <stdlib.h>
#include <stdio.h>
extern "C" {
#include <readline/readline.h>
#include <readline/history.h>
};
#endif

#include "error.h"
#include "syntax.h"
#include "symbols.h"
#include "quheap.h"
#include "options.h"

int isErrorReported=0;

void initialize_readline() {
#ifdef QCL_USE_READLINE
rl_bind_key ('\t', rl_insert);
#endif
}

void qclerror(string e,string m) {
  cerr << "! " << e << '\n';
  if(m!="") cerr << m << '\n';
  if(optLogfile) {
    *optLogfile << "! " << e << '\n';
  }
  isErrorReported=1;
}

void qclabort(string e,string m) { 
  if(e!="") qclerror(e,m);
  cerr << "! QCL Terminates on fatal error\n";
  if(optLogfile) {
    *optLogfile << "! QCL Terminates on fatal error\n";
    qcl_delete(optLogfile);
    optLogfile=0;
  }
  exit(1);
}

string qclinput(string p) {
  string t="";
#ifdef QCL_USE_READLINE
  char *l;
  l=readline((const char*)p);
  if(l) {
    t=l; t+='\n';
    if(*l) add_history(l);
    free(l);
  }
#else
  cout << p;
  readline(cin,t);
#endif
  isErrorReported=0;
  return t;
}
  
void qcloutput(string s) {
  cout << s;
}  

const char *ERROR_STR[] = {
  "no error","exit","message","internal error","illegal token","parse error",
  "type mismatch","invalid type","parameter mismatch","invalid parameter",
  "unknown symbol","illegal scope","invalid definition","math error",
  "runtime error","general error","syntax error","user error",
  "range error", "external error", "I/O-error", "shell error", 
  "memory error", "option error"
};

void tError::reset() {
  err=errNONE; lineno=0; linetxt=msg=sym=""; obj=sVOID;
  object=0; local=global=0; quheap=0; def=0; 
  isErrorReported=0; 
}

void tError::trace(sObject *obj,SymTable *loc,SymTable *gl,QuHeap *qh,sDef *d) {
  object=obj;
  local=loc;
  global=gl;
  quheap=qh;
  def=d;  
  
  if(object) cerr << " Object " << object->objstr();
  if(def) cerr << " (in " << def->objstr() << ")";
  if(object->isExpr()) cerr << " type " << ((sExpr*)object)->type().str();
  cerr << "\n";
  if(object) cerr << "  line:  " << object->prtstr() << "\n";
//if(global) cerr << "  global:" << global->prtstr() << "\n";
  if(local)  cerr << "  local: " << local->prtstr() << "\n";
  if(quheap) cerr << "  quheap:" << quheap->prtstr() << "\n";
}

string tError::errmsg() {
  string o;
  if(sym!="" && obj!=sVOID) {
    switch(obj)  {
      case sFUNCTDEF: o += "in function ";  break;
      case sPROCDEF:  o += "in procedure "; break;
      case sQUOPDEF:  o += "in operator ";  break;
      case sQUFUNDEF:  o += "in qufunct ";  break;
      default: ;
    };
    o += sym + ": ";
  };
  o += ERROR_STR[err];
  if(msg!="") o += ": " + msg;
  return o;
}


void tError::printerr(ostream& o) {
  if(lineno) o << "line " << lineno << " ";
  if(linetxt.length()>0) {
     o << linetxt << "\n  ";
  };
  if(sym!="" && obj!=sVOID) {
    switch(obj)  {
      case sFUNCTDEF: o << "in function ";  break;
      case sPROCDEF:  o << "in procedure "; break;
      case sQUOPDEF:  o << "in operator ";  break;
      case sQUFUNDEF:  o << "in qufunct ";  break;
      default: ;
    };
    o << sym << ": ";
  };
  o  << ERROR_STR[err];
  if(msg!="") o << ": " << msg;
  o << "\n";
}

void tError::report(ErrType e,sDef *d) {
  err=e;
  msg="";
  if(d) {
    sym=d->id();
    obj=d->object();
  } else{
    sym="";
    obj=sVOID;
  };
}

void tError::line(string txt,int l) {
  if(txt.length()>60) 
    linetxt=txt.at(0,56)+" ...";
  else
    linetxt=txt;
  lineno=l;
}

void tError::print(string s) {
  qcloutput(": "+s);
}

void tError::log(string s) {
  if(optLogfile) *optLogfile << s;
}

string tError::input(string s) {
  return qclinput("? "+s);
}

void tError::trBeginTypecheck(sObject *obj,SymTable *loc,SymTable *gl,QuHeap *qh,sDef *d) {
}
  
void tError::trEndTypecheck(sObject *obj,SymTable *loc,SymTable *gl,QuHeap *qh,sDef *d) {
}

void tError::trBeginExec(sObject *obj,SymTable *loc,SymTable *gl,QuHeap *qh,sDef *d) {
}

void tError::trEndExec(sObject *obj,SymTable *loc,SymTable *gl,QuHeap *qh,sDef *d) {
}

