/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/



#include "syntax.h"
#include "error.h"
#include "debug.h"
#include "options.h"

#include <stdlib.h>
#include <math.h>
#include <builtin.h>

#define TRACE() if (optTrace) { cerr << "EVAL"; err->trace(this,loc,gl,qh,def); }

#define EVERR(e,s) { 			\
  err->report(e,def);			\
  if(s) err->remark(s);			\
  if(1 || isStmt() || isDef()) 		\
    err->line(prtstr(),0);		\
  if(optDebug) qclshell(loc,gl,qh,def,err,DB_ERROR); \
  return tValue(tERROR);		\
}

#define ERR tValue(tERROR);
#define CHECK(v) if((v).isError()) return ERR
#define EVAL(v,p) { v=(p)->eval(loc,gl,qh,def,err); CHECK(v); }
#define ASSERT(c) if(!(c)) EVERR(errINT,"eval type inconsistency")

tValue sExpr::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  EVERR(errINT,"eval is not implemented on class sExpr");
}

tValue sConst::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  return _val;
}

tValue sVar::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  ObjType o;
  TRACE();
  v=loc->getVal(id(),&o);
  if(!v.isBaseType()) v=gl->getVal(id(),&o);
  ASSERT(v.type()==type() || (v.isQuExpr() && type().isQuExpr()));
  return v;
}

tValue sSubscript::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v,s;
  int i;
  quState *q;
  TRACE();
  EVAL(v,pvar);
  EVAL(s,psub);
  ASSERT(v.isQuExpr() && s.type()==tINTEGER);
  i=(int)s;  
  q=v.qustate();
  if(i<0 || i>=q->mapbits()) 
    EVERR(errRANGE,"invalid qubit subscript");
  return tValue(q->newsubstring(1,i));
}

tValue sSubRange::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v,e1,e2;
  int i,j,m;
  quState *q;
  TRACE();
  EVAL(v,pvar);
  EVAL(e1,pexpr1);
  EVAL(e2,pexpr2);
  ASSERT(v.isQuExpr() && e1.isInt() && e2.isInt());
  i=(int)e1;
  j=(int)e2;
  if(format==SUBRANGE_END) j=j-i+1; 
  q=v.qustate();
  m=q->mapbits(); 
  if(i<0 || j<1 || i+j>m)
    EVERR(errRANGE,"invalid quantum subregister");
  return tValue(q->newsubstring(j,i));
} 

tValue sNeg::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  TRACE();
  EVAL(v,pexpr);
  switch(v.type()) {
    case tINTEGER: 	return tValue(-(int)v);
    case tREAL: 	return tValue(-(tReal)v);
    case tCOMPLEX: 	return tValue(-(tComplex)v);
  };
  ASSERT(0);
}

tValue sNot::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  TRACE();
  EVAL(v,pexpr);
  ASSERT(v.type()==tBOOLEAN);
  return v==TRUE ? tValue(FALSE) : tValue(TRUE);
}

tValue sLength::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  TRACE();
  EVAL(v,pexpr);
  ASSERT(v.isQuExpr());
  return tValue((int)v.qustate()->mapbits());
}

tValue sBinOp::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v,e1,e2;
  int inv=0,i,n;
  TRACE();
  EVAL(e1,pexpr1);
  EVAL(e2,pexpr2);

  #define RET(t,o) return tValue((t)(((t)e1) o ((t)e2)));
  if(isArithOp()) {
    ASSERT(e1.isNumber() && e2.isNumber());
    ASSERT(object()!=sMOD || (e1.isInt() && e2.isInt()));
    if((object()==sMOD || object()==sDIV) && e2.isZero())
      EVERR(errMATH,"division by zero"); 
    if(object()==sPOW) {
      if(e1.isZero() && e2.isZero()) EVERR(errMATH,"0^0 is undefined");
      if(type().isInt() && ((int)e2)<0) EVERR(errMATH,"negative exponent in integer power");
      if(type().isReal() && ((tReal)e1)<0 && !e2.isInt())
        EVERR(errMATH,"negative base in non integer real value power");
    }
    if(type()==tINTEGER) 
      switch(object()) {
        case sADD:	RET(int,+);
        case sSUB:	RET(int,-);
        case sMULT:	RET(int,*);
        case sDIV:	RET(int,/);
        case sMOD:	RET(int,%);
        case sPOW:	return tValue((int)pow((long)(int)e1,(long)(int)e2));
        //n=1; for(i=0;i<(int)e2;i++) n*=(int)e1; return tValue((int)n);
    };
    if(type()==tREAL) 
      switch(object()) {
        case sADD:	RET(tReal,+);
        case sSUB:	RET(tReal,-);
        case sMULT:	RET(tReal,*);
        case sDIV:	RET(tReal,/);
        case sPOW:	return tValue((tReal)pow((tReal)e1,(tReal)e2));
    };
    if(type()==tCOMPLEX) 
      switch(object()) {
        case sADD:	RET(tComplex,+);
        case sSUB:	RET(tComplex,-);
        case sMULT:	RET(tComplex,*);
        case sDIV:	RET(tComplex,/);
        case sPOW:	return tValue((tComplex)pow((tComplex)e1,(tComplex)e2));
    };
    ASSERT(0);
  };
  #undef RET
  
  #define RET(t,o) return (t)e1 o (t)e2 ? tValue(TRUE) : tValue(FALSE)
  if(isCompOp()) {
    if(object()==sLEEQ || object()==sLESS) {
      ASSERT(e1.isOrdered() && e2.isOrdered()); 
      if(e1.isInt() && e2.isInt()) { 
        if(object()==sLEEQ) RET(int,<=); else RET(int,<);
      } else {  
        if(object()==sLEEQ) RET(tReal,<=); else RET(tReal,<);
      }
    } else {
      if(e1.isNumber() && e2.isNumber()) {
        if(e1.isInt() && e2.isInt()) { 
          if(object()==sEQUAL) RET(int,==); else RET(int,!=);
        };
        if(e1.isComplex() || e2.isComplex()) {
          if(object()==sEQUAL) RET(tComplex,==); else RET(tComplex,!=);
        };
        if(object()==sEQUAL) RET(tReal,==); else RET(tReal,!=);
      } else {
        ASSERT(e1.type()==e2.type());
        if(e1.isString()) {
          if(object()==sEQUAL) RET(string,==); else RET(string,!=);
        }
      }
    }
    ASSERT(0);
  }
  #undef RET
  
  #define RET(t,o) return (t)e1 o (t)e2 ? tValue(TRUE) : tValue(FALSE)
  if(isLogOp()) {
    ASSERT(e1.isBoolean() && e2.isBoolean());
    switch(object()) {
      case sAND: 	RET(boolean,&&);
      case sOR:		RET(boolean,||);
      case sXOR:	RET(boolean,!=);
    }
  }
  #undef RET

  if(object()==sCONCAT) { 
    ASSERT(e1.type()==e2.type());
    if(e1.isString()) return tValue((string)e1+(string)e2);
    if(e1.isQuExpr()) {
      bitvec v1=e1.qustate()->mapmask();
      bitvec v2=e2.qustate()->mapmask();
      if(!zero(v1&v2)) EVERR(errRANGE,"quantum registers overlap");
      return tValue(new quCombState(e1.qustate()->newclone(),e2.qustate()->newclone()));
    }
  }
  ASSERT(0);
}

tValue sFunctCall::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  SymTab sym;
  sArgDef ret(type(),new tId(RETURNID));
  sRoutDef *d;
  sExpr *p;
  sDef *q;
  void *cp,*cq;
  int r;
  
  TRACE();
  d=(sRoutDef*)gl->getDef(id());
  ASSERT(d && d->isFunctDef());
  sym.put(&ret);
  for_pplist(p,ppar,cp,q,d->args(),cq,p && q) {
    EVAL(v,p);
    if(sym.put(q,v)) EVERR(errINT,"argument binding failed");
  }
  if(d->invoke(&sym,gl,qh,err)) return ERR;
  v=sym.getVal(RETURNID);
  if(v.isUndef()) EVERR(errGEN,"function didn't return a value");
  if(!v.isBaseType()) EVERR(errINT,"invalid return value"); 
  return v;
}



tValue sBaseFunct::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  int n;
  tReal r;
  tComplex z;
  TRACE();
  EVAL(v,pexpr);
  if(v.isInt()) {
    n=(int)v;
    r=(tReal)v;
    z=(tComplex)v;
  }
  if(v.isReal()) {
    r=(tReal)v;
    z=(tComplex)v;
  }    
  if(v.isComplex()) {
    z=(tComplex)v;
  }
  if(type().isInt()) {
    ASSERT(v.isNumber());
    switch(object()) {
      case sFLOOR  :  return tValue((int)floor(r));
      case sCEIL   :  return tValue((int)ceil(r));
      case sABS    :  return tValue((int) n>=0 ? n : -n);

    }
  };
  if(type().isReal()) {
    ASSERT(v.isNumber());
    switch(object()) {
      case sEXP    :  return tValue((tReal)exp(r));
      case sSIN    :  return tValue((tReal)sin(r));
      case sCOS    :  return tValue((tReal)cos(r));
      case sTAN    :  return tValue((tReal)tan(r));
      case sCOT    :  return tValue((tReal)1./tan(r));
      case sSINH   :  return tValue((tReal)sinh(r));
      case sCOSH   :  return tValue((tReal)cosh(r));
      case sTANH   :  return tValue((tReal)tanh(r));
      case sCOTH   :  return tValue((tReal)1./tanh(r));
      case sABS    :  return tValue((tReal)abs(z));
      case sRE     :  return tValue((tReal)z.real());
      case sIM     :  return tValue((tReal)z.imag());
      case sCONJ   :  return v;
      case sSQRT   :
        if(r<0) EVERR(errMATH,"real square root of negative number");  
        return tValue((tReal)sqrt(r));
    }
  };
  if(type().isComplex()) {
    ASSERT(v.isNumber());
    switch(object()) {
      case sEXP    :  return tValue((tComplex)exp(z));
      case sSIN    :  return tValue((tComplex)sin(z));
      case sCOS    :  return tValue((tComplex)cos(z));
      case sTAN    :  return tValue((tComplex)sin(z)/cos(z));
      case sCOT    :  return tValue((tComplex)cos(z)/sin(z));
      case sSINH   :  return tValue((tComplex)sinh(z));
      case sCOSH   :  return tValue((tComplex)cosh(z));
      case sTANH   :  return tValue((tComplex)sinh(z)/cosh(z));
      case sCOTH   :  return tValue((tComplex)cosh(z)/sinh(z));
      case sCONJ   :  return tValue((tComplex)conj(z));
      case sSQRT   :  return tValue((tComplex)sqrt(z));
    }
  };
  ASSERT(0);
}

tValue sListFunct::eval(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  sExpr *p;
  void *c;
  int i;
  int argc=pexpr->n();
  tValue v[argc];
  int n[argc];
  tReal r[argc];
  tComplex z[argc];
  int rn;
  tReal rr;
  tComplex rz;
  TRACE();
  i=0;
  for_plist(p,pexpr,c) {
    EVAL(v[i],p);
    if(v[i].isInt()) {
      n[i]=(int)v[i];
      r[i]=(tReal)v[i];
      z[i]=(tComplex)v[i];
    }
    if(v[i].isReal()) {
      r[i]=(tReal)v[i];
      z[i]=(tComplex)v[i];
    }
    if(v[i].isComplex()) {
      z[i]=(tComplex)v[i];
    }
    i++;
  }
  switch(object()) {
    case sLOG:
      ASSERT(v[0].isNumber() && (!pexpr2 || v[1].isNumber()));
      if(type().isComplex()) {
        if(pexpr2)
          return tValue((tComplex)log(z[0])/log(z[1]));
        else
          return tValue((tComplex)log(z[0]));
      } else {
        if(r[0]<=0 || (pexpr2 && r[1]<=0))
	  EVERR(errMATH,"real logarithm of non positive number");
        if(pexpr2)
          return tValue((tReal)log(r[0])/log(r[1]));
        else
          return tValue((tReal)log(r[0]));
      }
      break;
    case sRANDOM:
      ASSERT(argc==0 && (!def || def->isProcDef()));
      return tValue((tReal)random()/(tReal)(RAND_MAX+1.0));
    case sMIN:
    case sMAX:
      ASSERT(argc>0);
      if(type().isInt()) {
        rn=n[0];
        for(i=1;i<argc;i++) {
          if(object()==sMIN && rn>n[i]) rn=n[i];
          if(object()==sMAX && rn<n[i]) rn=n[i];
        }
        return tValue((int)rn);
      } else {
        rr=r[0];
        for(i=1;i<argc;i++) {
          if(object()==sMIN && rr>r[i]) rr=r[i];
          if(object()==sMAX && rr<r[i]) rr=r[i];
        }
        return tValue((tReal)rr);
      }
      break;
    case sGCD:
    case sLCM:
      ASSERT(argc>0);
      rn=n[0];
      for(i=1;i<argc;i++) {
        if(object()==sGCD) rn=gcd(n[i],rn);
        if(object()==sLCM) rn=lcm(n[i],rn);
      }
      return tValue((int)rn);      
  }
  ASSERT(0);  
}


