/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/



#include "syntax.h"
#include "error.h"
#include "quheap.h"
#include "options.h"
#include "debug.h"

#define TRACE() if (optTrace) { cerr << "EXEC";err->trace(this,loc,gl,qh,def); }

#define EXERR(e,s) { 		\
  err->report(e,def);			\
  if(s) err->remark(s);			\
  if(1 || isStmt() || isDef()) 		\
    err->line(prtstr(),0);		\
  if(optDebug) qclshell(loc,gl,qh,def,err,DB_ERROR); \
  return 1;				\
}

#define ERR 1
#define RET 2
#define CHECK(v) { if((v).isError()) return ERR; }
#define EVAL(v,p) { if(p) { v=(p)->eval(loc,gl,qh,def,err); CHECK(v); } }

/***********************************************/
/***************** Definitions *****************/
/***********************************************/


#define ASSERT(c) if(!(c)) EXERR(errINT,"define type inconsistency")

int sDef::define(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  return ERR;
}

int sDefList::define(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  sDef *p;
  void *c;
  TRACE();
  for_plist(p,this,c) if(p->define(loc,gl,qh,def,err)) return 1;
  return 0;
}

int sVarDef::define(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v,l;
  int e;
  TRACE();
  if(pinit) {
    EVAL(v,pinit);
    ASSERT(type()==v.type());
  } else {
    if(type().isQuExpr()) {
      ASSERT(plen);
      EVAL(l,plen);
      if((int)l<=0) 
        EXERR(errMEM,"quantum register of non-positive length");
      v=qh->qualloc((int)l,(BaseType)type());
      if(v.isError()) EXERR(errMEM,"not enough quantum memory");
    } else {
      v=tValue(type());
    }
  }
  if(def) e=loc->put(this,v); else e=gl->put(this,v);
  if(e) EXERR(errINT,"cannot store variable");
  return 0;
}

int sConstDef::define(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  int e;
  TRACE();
  EVAL(v,pexpr);
  ASSERT(type()==v.type());
  if(def) e=loc->put(this,v); else e=gl->put(this,v);
  if(e) EXERR(errINT,"cannot store constant");
  return 0;
}

int sRoutDef::define(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  ASSERT(!def);
  if(gl->put(this)) EXERR(errINT,"cannot store routine");
  return 0;
}

#undef TRACE
#define TRACE() if (err->isTrace()) err->trace(this,loc,gl,qh,def)


/***********************************************/
/***************** Statements ******************/
/***********************************************/


#undef ASSERT
#define ASSERT(c) if(!(c)) EXERR(errINT,"exec type inconsistency")

int sStmt::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  return ERR;
}

int sStmtList::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  sStmt *p;
  void *c;
  int r;
  TRACE();
  for_plist(p,this,c) {
    r=p->exec(loc,gl,qh,def,err);
    if(r) return r;
  };
  return 0;
}

int sCall::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  SymTab *psym;
  sRoutDef *d;
  sExpr *p;
  sDef *q;
  void *cp,*cq;
  int r;
  bitvec m(qh->nBits());
  TRACE();
  d=(sRoutDef*)gl->getDef(id());
  ASSERT(d && (d->isProcDef() || d->isQuDef()));
  psym=new SymTab();
  for_pplist(p,ppar,cp,q,d->args(),cq,p && q) {
    v=(p)->eval(loc,gl,qh,def,err);
    if(v.isError()) { qcl_delete(psym); return ERR; }
    if(psym->put(q,v)) { 
      qcl_delete(psym); 
      EXERR(errINT,"argument binding failed");
    }
    if(v.isQuExpr()) {
      if(zero(m & v.qustate()->mapmask())) {
        m |= v.qustate()->mapmask();
      } else {
        qcl_delete(psym);
        EXERR(errRUN,"quantum arguments overlapping");
      }
    }
  }
  if(d->isQuDef() && qh->isDelayed()) {
    r=qh->add(d,psym,isInv());
  } else {
    r=d->invoke(psym,gl,qh,err,isInv());
    qcl_delete(psym);
  }
  return r;
}

int sAssign::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue *p;
  tValue v;
  TRACE();
  ASSERT((p=loc->getRef(pvar->id())) || (p=gl->getRef(pvar->id())));
  EVAL(v,pexpr);
  *p=v;
  return 0;
}

int sTrans::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  string opid= object()==sFANOUT ? FANOUTID : SWAPID;
  tValue a,b;
  sRoutDef *fo;
  SymTab *psym;
  void *c;
  int r;
  TRACE();
  EVAL(a,plval);
  EVAL(b,prval);
  ASSERT(a.isQuExpr() && b.isQuExpr());
  if(a.qustate()->mapbits() != b.qustate()->mapbits())
    EXERR(errRUN,opid+" arguments are of different length");
  switch(object()) {
    case sFANOUT:
      if(qh->isDelayed()) {
    	r=qh->addFanout(gl,err,a,b,isInv());
      } else {
    	r=qh->applyFanout(gl,err,a,b,isInv());
      }
      break;
    case sSWAP:
      if(qh->isDelayed()) {
    	r=qh->addSwap(gl,err,a,b,isInv());
      } else {
    	r=qh->applySwap(gl,err,a,b,isInv());
      }
      break;
    default: EXERR(errINT,"undefined sTrans object");
  }
  return r;
}

int sFor::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue *p;
  tValue f,t,s;
  int r,i,st;
  TRACE();
  ASSERT((p=loc->getRef(pvar->id())) || (p=gl->getRef(pvar->id())));
  EVAL(f,pfrom);
  EVAL(t,pto);
  if(pstep) EVAL(s,pstep);
  ASSERT(f.isInt() && t.isInt() && (!pstep || s.isInt()));
  if(pstep) {
    st=(int)s;
  } else {
    st= 1;
  }
/*
  if(((int)f>(int)t && st>0) || ((int)f<(int)t && st<0)) return 0;
  if(!st || ((int)t-(int)f>0 && st<=0) || ((int)t-(int)f<0 && st>=0))
    EXERR(errRUN,"infinite for loop"); 
*/

  if(!st) EXERR(errRUN,"zero increment in for loop"); 
  for(i=(int)f;(i<=(int)t && st>=0) || (i>=(int)t && st<=0);i+=st) {
    *p=tValue(i);
    r=pbody->exec(loc,gl,qh,def,err);
    if(r) return r;
  }
  return 0;
}

int sIf::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  TRACE();
  EVAL(v,pcond);
  ASSERT(v.isBoolean());
  if((boolean)v==TRUE) {
    return pthen->exec(loc,gl,qh,def,err);
  } else {
    if(pelse) return pelse->exec(loc,gl,qh,def,err);
  }
  return 0;
}

int sLoop::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  int r;
  TRACE();
  while(1) {
    if(!isUntil) {
      EVAL(v,pcond);
      ASSERT(v.isBoolean());
      if((boolean)v==FALSE) return 0;
    }
    r=pbody->exec(loc,gl,qh,def,err);
    if(r) return r;
    if(isUntil) {
      EVAL(v,pcond);
      ASSERT(v.isBoolean());
      if((boolean)v==TRUE) return 0;
    }
  }
  return ERR;
}

int sReturn::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue *p;
  tValue v;
  TRACE();
  EVAL(v,pexpr);
  p=loc->getRef(RETURNID);
  ASSERT(p);
  *p=v;
  return RET;
}

int sInput::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  int n,e;
  tReal x,y;
  tValue v;
  tValue *p;
  string m,s;
  TRACE();
  ASSERT( (p=loc->getRef(pvar->id())) || (p=gl->getRef(pvar->id())) );
  if(pprompt) {
    EVAL(v,pprompt);
    m=(string)v;
  } else {
    m+=pvar->type().str()+" "+pvar->id()+" ";
    if(pvar->type().isComplex()) m+="[(Re,Im)] ";
    if(pvar->type().isBoolean()) m+="[t(rue)/f(alse)] ";
    m+="?";
  }
  m+=" ";
  v=tValue(tERROR);
  while(1) {
    e=0;
    s=err->input(m)+"\n";
    switch(pvar->type()) {
      case tINTEGER:
        if(sscanf((const char*)s,"%d",&n)==1) v=tValue((int)n);
        break;
      case tREAL:
        if(sscanf((const char*)s,"%lf",&x)==1) v=tValue((tReal)x);
        break;
      case tCOMPLEX:
        if(sscanf((const char*)s,"(%lf,%lf)",&x,&y)==2) v=tValue(tComplex(x,y));
        break;
      case tSTRING:
        v=tValue(s);
        break;
      case tBOOLEAN:
        if(s[0]=='t' || s[0]=='T')  v=tValue((boolean)TRUE);
        if(s[0]=='f' || s[0]=='F')  v=tValue((boolean)FALSE);
        break;
      default: ASSERT(0);
    };
    if(v.isDefined()) break;
  }
  *p=v;
  err->log("? "+m+s);
  return 0;
}

int sPrint::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  sExpr *p;
  void *c;
  string ostr;
  tValue v;
  TRACE();
  for_plist(p,plist,c) {
    EVAL(v,p);
    ostr+=v.str()+" ";
  }
  ostr+="\n";
  err->print(ostr);
  err->log(": "+ostr);
  return 0;
}

int sExit::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue v;
  TRACE();
  if(pmsg) {
    EVAL(v,pmsg);
    EXERR(errUSR,(string)v);
  } else {
    err->report(errEXIT,def);
    err->line(prtstr(),0);
    return 1;
  }
}

int sMeasure::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  tValue s,m;
  tValue *p;
  ostream *log=&cout;
  TRACE();
  EVAL(s,pstate);
  ASSERT(s.isQuExpr());
  if(pvar) ASSERT(((p=loc->getRef(pvar->id())) || (p=gl->getRef(pvar->id()))) && p->isInt());
  if(optLog) {
    if(optLogfile) log=optLogfile;
    *log << "@ MEASURE " << s << "\n";
  };
  m=qh->measure(s);
  isStateModified=1;
  if(m.isUndef()) EXERR(errMATH,"Measured Integer is too long");
  if(!m.isInt()) EXERR(errINT,"Measurement failed");
  if(pvar) *p=m;
  return 0;
}

int sInclude::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  EXERR(errINT,"uncaught include");
  return 1;
}


  
int sReset::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  TRACE();
  if(optLog) {
    if(optLogfile) *optLogfile << "@ RESET\n"; else cout << "@ RESET\n";
  };
  if(qh->reset()) EXERR(errINT,"reset failed");
  isStateModified=1;
  return 0;
}


/***********************************************/
/****************** Commands *******************/
/***********************************************/


int sDump::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  string s;
  char t[80];
  tValue v;
  int i,j;
  quState *q;
  bitvec b(qh->nBits()),c;
  probtree *p;
  TRACE();

  bitvec_format=BITVEC_FORMAT_PLAIN;
  if(optDumpFormat=='x') bitvec_format|=BITVEC_FORMAT_PLAIN;  
  if(pexpr) {
    EVAL(v,pexpr);
    ASSERT(v.isQuExpr());
    q=v.qustate();
    ASSERT(q);
    s="SPECTRUM "+ pexpr->prtstr()+": "+v.str()+"\n";
/*
    s="SPECTRUM "+ pexpr->prtstr()+": "+v.str()+" Pos: |";
    for(i=qh->nBits()-1;i>=0;i--) {
      b.setbit(i,1);
      c=q->map(b);
      if(zero(c)) { 
        s+=".";
      } else {
        for(j=0;j<q->mapbits();j++) {
          if(c[j]) { s+=(char)(j<10 ? '0'+j : 'a'+j-10); break; }
        }
      }
      b.setbit(i,0); 
    }
    s+=">\n";
*/
    err->print(s);
    bitvec_format=BITVEC_FORMAT_PLAIN;
    if(optDumpFormat=='a' && q->mapbits()>32 || optDumpFormat=='x') 
      bitvec_format|=BITVEC_FORMAT_HEX;
    cout << (*q) << "\n";
  } else {
    term_format=TERM_FORMAT_ALGEBRAIC;
    bitvec_format=BITVEC_FORMAT_BIN|BITVEC_FORMAT_KET;
    if(optDumpFormat=='a' && qh->nBits()>32 || optDumpFormat=='x') 
      bitvec_format|=BITVEC_FORMAT_HEX;
    sprintf(t,"STATE: %d / %d qubits allocated, %d / %d qubits free\n",
      qh->nBits()-qh->nFree(),qh->nBits(),qh->nFree(),qh->nBits());
    s=t;
    err->print(s);
    printbase(cout,*qh->state(),0,"\n");
  }            
  return 0;
}

int sInspect::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  sExpr *p;
  sVar *q;
  void *c;
  sDef *d;
  string ostr;
  tValue *pv=0;
  TRACE();
  if(plist) {
    for_plist(p,plist,c) {
      ASSERT(p->object()==sVAR);
      q=(sVar*)p;
      if(d=loc->getDef(q->id())) {
  	ostr="local ";
	pv=loc->getRef(q->id());
      } else if(d=gl->getDef(q->id())) {
  	ostr="global ";
	pv=gl->getRef(q->id());
       }
      ostr+="symbol "+q->id();
      if(pv) ostr+=" = "+pv->str();
      if(d) {
  	ostr+=":\n"+d->prtstr(INDENT);
	if(d->isArgDef()) ostr+="\n";
      } else {
  	ostr+=" is undefined.\n";
      };
      err->print(ostr);
    };
  } else {
    err->print("global symbols:\n"+gl->prtstr(INDENT));
    err->print("local symbols:\n"+loc->prtstr(INDENT));
  };
  return 0;
}

int sLoad::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  ifstream f;
  tValue v;
  string fn;
  TRACE();
  if(pexpr) {
    EVAL(v,pexpr);
    ASSERT(v.isString());
    fn=(string)v;
  } else {
    fn=DEF_LOAD_FILE DEF_DATAFILE_EXT;
  };
  f.open(fn,ios::in);
  if(!f) EXERR(errIO,"Can't open "+fn+" for reading");
  if(qh->load(f)) EXERR(errIO,"Error while reading "+fn);
  f.close();
  return 0;
}

int sSave::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  ofstream f;
  tValue v;
  string fn;
  TRACE();
  if(pexpr) {
    EVAL(v,pexpr);
    ASSERT(v.isString());
    fn=(string)v;
  } else {
    fn=DEF_SAVE_FILE DEF_DATAFILE_EXT;
  };
  f.open(fn,ios::out);
  if(!f) EXERR(errIO,"Can't open "+fn+" for writing");
  if(qh->save(f)) EXERR(errIO,"Error writing reading "+fn);
  f.close();
  if(!f) EXERR(errIO,"Can't close "+fn);
  return 0;
}

int sSet::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  int i;
  tValue v;
  string s;
  TRACE();
  EVAL(v,pexpr);
  for(i=0;qcl_options[i].val;i++) {
    if(opt==string(qcl_options[i].name)) goto cont;
  }
  EXERR(errOPT,"illegal option "+opt);
cont:
  if(qcl_options[i].has_arg && !pexpr)
    EXERR(errOPT,"missing argument for option "+opt);
  if(!qcl_options[i].has_arg && pexpr)
    EXERR(errOPT,"option "+opt+" takes no argument");
  if(pexpr) s=v.str();
  if(evalopt(qcl_options[i].val,s,def,err)) return 1;
  return 0;
}

int sShell::exec(SymTable *loc,SymTable *gl,QuHeap *qh,sDef *def,tError *err) {
  string ostr;
  TRACE();
  if(!optShellEscape) return 0;
  qclshell(loc,gl,qh,def,err,DB_ESCAPE);
  return 0;
}

 
