/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#pragma implementation

#include "extern.h"
#include "syntax.h"
#include "error.h"
#include "operator.h"
#include "quheap.h"
#include "options.h"

#define ROUTINE(n) int n(SymTable *loc,SymTable *gl,QuHeap *qh,tError *err,int inv)
#define EXTERR(s) { err->report(errEXT); if(s) err->remark(s); return 1; }
#define EXTASSERT(c) { if(!(c)) EXTERR("Assert failed"); }
#define GETVAL(s) tValue *value; value=loc->getRef(s); // cerr << s << " = " << *value << "\n";
#define PAR_INT(v,s) int v; { GETVAL(s); if(!value->isInt()) EXTERR("integer expected"); v=(int)(*value); }
#define PAR_REAL(v,s) tReal v; { GETVAL(s); if(!value->isReal()) EXTERR("real expected"); v=(tReal)(*value); }
#define PAR_COMPLEX(v,s) tComplex v; { GETVAL(s); if(!value->isComplex()) EXTERR("complex expected"); v=(tComplex)(*value); }
#define PAR_QUSTATE(v,s) quState* v; { GETVAL(s); if(!value->isQuExpr()) EXTERR("quantum expression expected"); v=value->qustate(); }
#define STOP while(*((int*)0));

//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct Fanout(quconst a,quvoid b);

class opFANOUT : public opGate {
  int inv;
public:
  opFANOUT(int n,int i) : opGate(2*n) { inv=i; }
  virtual term funct(const bitvec& v) const { 
    bitvec a(v.getbits(0,bits()/2));
    bitvec b(v.getbits(bits()/2,bits()/2));
    b^=a;
    a+=b;
    return term(a,1.0);
  };
  virtual opOperator *newclone() const { return new opFANOUT(bits()/2,inv); }
};

ROUTINE(ext_fanout) {
  PAR_QUSTATE(a,"a");
  PAR_QUSTATE(b,"b");
  int na,nb;
  bitvec ma,mb;
  na=a->mapbits();
  ma=a->mapmask();
  nb=b->mapbits();
  mb=b->mapmask();
  if(na!=nb || !zero(ma & mb)) 
    EXTERR("fanout arguments must be of equal size")
  quCombState cb(a,b); 
  opFANOUT(na,inv).apply(cb);
  return 0;
}
  
//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct Swap(qureg a,qureg b);

ROUTINE(ext_swap) {
  PAR_QUSTATE(sa,"a");
  PAR_QUSTATE(sb,"b");
  int na,nb;
  bitvec ma,mb;
  na=sa->mapbits();
  ma=sa->mapmask();
  nb=sb->mapbits();
  mb=sb->mapmask();
  if(na!=nb || !zero(ma & mb)) 
    EXTERR("swap arguments must be of equal size")
  quCombState cb(sa,sb); 
  opSwap(na+nb,na,0,na).apply(cb);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern operator Matrix2x2(
//!   complex u00,complex u01,
//!   complex u10,complex u11,
//! qureg q);
//!
//! extern operator Matrix4x4(
//!   complex u00,complex u01,complex u02,complex u03,
//!   complex u10,complex u11,complex u12,complex u13,
//!   complex u20,complex u21,complex u22,complex u23,
//!   complex u30,complex u31,complex u32,complex u33,
//! qureg q);
//!
//! extern operator Matrix8x8(
//!   complex u00,complex u01,complex u02,complex u03,
//!   complex u04,complex u05,complex u06,complex u07,
//!   complex u10,complex u11,complex u12,complex u13,
//!   complex u14,complex u15,complex u16,complex u17,
//!   complex u20,complex u21,complex u22,complex u23,
//!   complex u24,complex u25,complex u26,complex u27,
//!   complex u30,complex u31,complex u32,complex u33,
//!   complex u34,complex u35,complex u36,complex u37,
//!   complex u40,complex u41,complex u42,complex u43,
//!   complex u44,complex u45,complex u46,complex u47,
//!   complex u50,complex u51,complex u52,complex u53,
//!   complex u54,complex u55,complex u56,complex u57,
//!   complex u60,complex u61,complex u62,complex u63,
//!   complex u64,complex u65,complex u66,complex u67,
//!   complex u70,complex u71,complex u72,complex u73,
//!   complex u74,complex u75,complex u76,complex u77,
//! qureg q);

ROUTINE(ext_matrix) {
  int i,j,k,n,dim;
  PAR_QUSTATE(q,"q");
  tValue v;
  tComplex z;
  
  n=q->mapbits();
  dim=(1 << n);
  if(!loc->getRef(string("u")+dec(dim-1)+dec(dim-1)) || loc->getRef(string("u")+dec(dim)+dec(dim)))
    EXTERR("illegal register size for matrix operator");
  tComplex u[dim][dim];
  term* t[dim];
  for(i=0;i<dim;i++) for(j=0;j<dim;j++) {
    v=loc->getVal(string("u")+dec(j)+dec(i));
    if(!v.isComplex()) EXTERR("invalid matrix element");
    if(inv) u[j][i]=conj((tComplex)v); else u[i][j]=v;
  }
  for(i=0;i<dim;i++) for(j=0;j<dim;j++) {
    z=0;
    for(k=0;k<dim;k++) z+=u[i][k]*conj(u[k][j]);
    if(abs(z-(i==j))>EPSILON_MATRIX)
      EXTERR("matrix operator is not unitary");
  }  
  for(i=0;i<dim;i++) {
    k=0;
    t[i]=new term[dim+1];
    for(j=0;j<dim;j++) {
      if(u[i][j]==0) continue;
      t[i][k++]=term(bitvec(n,j),u[i][j]);
    }
  }
  opMatrix(n,&t[0]).apply(*q);
  for(i=0;i<dim;i++) qcl_delarray(t[i]);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct Perm2(int p0 ,int p1 ,qureg q);
//!
//! extern qufunct Perm4(int p0 ,int p1 ,int p2 ,int p3 ,qureg q);
//!
//! extern qufunct Perm8(
//!   int p0 ,int p1 ,int p2 ,int p3 ,int p4 ,int p5 ,int p6 ,int p7 ,
//! qureg q);
//!
//! extern qufunct Perm16(
//!   int p0 ,int p1 ,int p2 ,int p3 ,int p4 ,int p5 ,int p6 ,int p7 ,
//!   int p8 ,int p9 ,int p10,int p11,int p12,int p13,int p14,int p15,
//! qureg q);
//!
//! extern qufunct Perm32(
//!   int p0 ,int p1 ,int p2 ,int p3 ,int p4 ,int p5 ,int p6 ,int p7 ,
//!   int p8 ,int p9 ,int p10,int p11,int p12,int p13,int p14,int p15,
//!   int p16,int p17,int p18,int p19,int p20,int p21,int p22,int p23,
//!   int p24,int p25,int p26,int p27,int p28,int p29,int p30,int p31,
//! qureg q);
//!
//! extern qufunct Perm64(
//!   int p0 ,int p1 ,int p2 ,int p3 ,int p4 ,int p5 ,int p6 ,int p7 ,
//!   int p8 ,int p9 ,int p10,int p11,int p12,int p13,int p14,int p15,
//!   int p16,int p17,int p18,int p19,int p20,int p21,int p22,int p23,
//!   int p24,int p25,int p26,int p27,int p28,int p29,int p30,int p31,
//!   int p32,int p33,int p34,int p35,int p36,int p37,int p38,int p39,
//!   int p40,int p41,int p42,int p43,int p44,int p45,int p46,int p47,
//!   int p48,int p49,int p50,int p51,int p52,int p53,int p54,int p55,
//!   int p56,int p57,int p58,int p59,int p60,int p61,int p62,int p63,
//! qureg q);

ROUTINE(ext_perm) {
  int i,n,dim;
  PAR_QUSTATE(q,"q");
  tValue v;
  
  n=q->mapbits();
  dim=(1 << n);
  if(!loc->getRef(string("p")+dec(dim-1)) || loc->getRef(string("p")+dec(dim)))
    EXTERR("illegal register size for permutation operator");
  int p[dim];
  int pp[dim];
  term t[dim];
  for(i=0;i<dim;i++) {
    v=loc->getVal(string("p")+dec(i));
    if(!v.isInt() || (int)v < 0 || (int)v >= dim) 
      EXTERR("invalid permutation element");
    p[i]=(int)v;
  }
  for(i=0;i<dim;i++) pp[i]=-1;
  for(i=0;i<dim;i++) pp[p[i]]=i;
  for(i=0;i<dim;i++) {
    if(pp[i]==-1) EXTERR("no permutation"); 
    if(inv) { 
      t[i]=term(bitvec(n,pp[i]),1); 
    } else {
      t[i]=term(bitvec(n,p[i]),1);
    }
  }
  opPermutation(n,t).apply(*q);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct Not(qureg q);

class opNOT : public opGate {
  int inv;
public:
  opNOT(int n,int i ) : opGate(2*n) { inv=i; }
  virtual term funct(const bitvec& v) const { return term(~v,1.0); };
  virtual opOperator *newclone() const { return new opNOT(bits()/2,inv); }
};

ROUTINE(ext_not) {
  PAR_QUSTATE(q,"q");
  opNOT(q->mapbits(),inv).apply(*q);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct CNot(qureg q,quconst c);

class opCNOT : public opGate {
  int inv;
  int arg;
  int ctr;
public:
  opCNOT(int a,int c,int i ) : opGate(a+c) { inv=i; arg=a; ctr=c; }
  virtual term funct(const bitvec& v) const {
    bitvec a(v.getbits(0,arg)); 
    bitvec c(v.getbits(arg,ctr));
    if(c==~bitvec(ctr)) a.not();
    return term(a+c,1.0); 
  };
  virtual opOperator *newclone() const { return new opCNOT(arg,ctr,inv); }
};

ROUTINE(ext_cnot) {
  PAR_QUSTATE(q,"q");
  PAR_QUSTATE(c,"c");
  quCombState cb(q,c);
  opCNOT(q->mapbits(),c->mapbits(),inv).apply(cb);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern operator CPhase(real phi,qureg q);

ROUTINE(ext_cphase) {
  PAR_REAL(phi,"phi");
  PAR_QUSTATE(q,"q");
  if(inv) phi=-phi;
  opCondPhase(q->mapbits(),q->mapbits(),phi).apply(*q);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern operator Rot(real theta,qureg q);

ROUTINE(ext_rot) {
  PAR_REAL(theta,"theta");
  PAR_QUSTATE(q,"q");
  if(q->mapbits()!=1) EXTERR("Only single qubits can be rotated");
  if(inv) theta=-theta;
  opU2(theta).apply(*q);
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern operator Mix(qureg q);

ROUTINE(ext_mix) {
  int i;
  quState *qb;
  PAR_QUSTATE(q,"q");
  for(i=0;i<q->mapbits();i++) {
    qb=q->newsubstring(1,i);
    opU2(PI/2,PI/2,PI/2,PI).apply(*qb);
    qcl_delete(qb);
  };
  return 0;
}

//////////////////////////////////////////////////////////////////////
//!
//! extern qufunct ModExp(int n,int x,quconst a,quvoid b);

ROUTINE(ext_modexp) {
  PAR_INT(n,"n");
  PAR_INT(x,"x");
  PAR_QUSTATE(a,"a");
  PAR_QUSTATE(b,"b");
  quCombState cb(a,b); 
  opEXPN(a->mapbits(),b->mapbits(),x,n).apply(cb);
  return 0;
}

RoutTableEntry ExtRoutTable[] = {
 { &ext_fanout,FANOUTID },
 { &ext_swap,SWAPID },
 { &ext_matrix,"Matrix2x2" },
 { &ext_matrix,"Matrix4x4" },
 { &ext_matrix,"Matrix8x8" },
 { &ext_perm,"Perm2" },
 { &ext_perm,"Perm4" },
 { &ext_perm,"Perm8" },
 { &ext_perm,"Perm16" },
 { &ext_perm,"Perm32" },
 { &ext_perm,"Perm64" },
 { &ext_not,"Not" },
 { &ext_cnot,"CNot" },
 { &ext_cphase,"CPhase" },
 { &ext_rot,"Rot" },
 { &ext_mix,"Mix" },
 { &ext_modexp,"ModExp" },
 { 0, "" }
};

//////////////////////////////////////////////////////////////////////
//!
//! boolean bit(int n,int b) {
//!   return n/2^b mod 2 == 1;
//! }
//!
//! qufunct set(int n,qureg q) {
//!   int i;
//!   for i=0 to #q-1 {
//!     if bit(n,i) { Not(q[i]); }
//!   }
//! }
//!
//! const pi=3.141592653589793238462643383279502884197;
