/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#include <stdio.h>
#include <time.h>
#include <string.h>

#include "options.h"
#include "error.h"

#define QCL_SHORTOPTS "hVino:b:s:I:p:f:d:a:l:L:c:T:S:E:t:e:r:"

struct option qcl_options[]={
  {"help",0,0,'h'},
  {"version",0,0,'V'},
  {"interactive",0,0,'i'},
  {"no-default-include",0,0,'n'},
  {"logfile",1,0,'o'},
  {"bits",1,0,'b'},
  {"seed",1,0,'s'},
  {"include-prefix",1,0,'I'},
  {"dump-prefix",1,0,'p'},
  {"dump-format",1,0,'f'},
  {"debug",1,0,'d'},
  {"auto-dump",1,0,'a'},
  {"log",1,0,'l'},
  {"log-state",1,0,'L'},
  {"check",1,0,'c'},
  {"trace",1,0,'T'},
  {"syntax",1,0,'S'},
  {"echo",1,0,'E'},
  {"test",1,0,'t'},
  {"shell-escape",1,0,'e'},
  {"allow-redefines",1,0,'r'},
//  {"",0,0,''},
  {0,0,0,0}
};

void printusage(char *name) {
  cerr << "USAGE: " << name << " [options] [file] ...\n";
  cerr << "QCL interpeter and quantum computer simulator\n\n";
  cerr << "Startup Options:\n";
  cerr << "-h, --help                      display this message\n";
  cerr << "-V, --version                   display version information\n";
  cerr << "-i, --interactive               force interactive mode\n";
  cerr << "-n, --no-default-include        don't read " << DEF_STD_INCLUDE << " on startup\n";
  cerr << "-o, --logfile                   specify a logfile\n";
  cerr << "-b, --bits=n:                   set number of qubits (" << 8*sizeof(int) <<")\n";
  cerr << "Dynamic Options (can be changed with the set statement):\n";
  cerr << "-s, --seed=<seed-value>         set random seed value (system time)\n";
  cerr << "-I, --include-path=<path>       QCL include path (.)\n";
  cerr << "-p, --dump-prefix=<file-prefix> set dump-file prefix\n";
  cerr << "-f, --dump-format=b,a,x         list base vectors as hex, auto or binary (a)\n";
  cerr << "-d, --debug=<y|n>               open debug-shell on error (n)\n";
  cerr << "-a, --auto-dump=<y|n>           allways dump quantum state in shell mode (n)\n"; 
  cerr << "-l, --log==<y|n>                log external operator calls (n)\n";
  cerr << "-L, --log-state==<y|n>          log state after each transformation (n)\n";
  cerr << "-c, --check==<y|n>              check consistency of quantum heap (n)\n";
  cerr << "-T, --trace==<y|n>              trace mode (very verbose) (n)\n";
  cerr << "-S, --syntax=<y|n>              check only the syntax, no interpretation (n)\n";
  cerr << "-E, --echo=<y|n>                echo parsed input (n)\n";
  cerr << "-t, --test=<y|n>                test programm, ignore quantum operations (n)\n";
  cerr << "-e, --shell-escape=<y|n>        honor shell-escapes (y)\n";
  cerr << "-r, --allow-redefines=<y|n>     ignore redefines instead of error (n)\n";
}

int optInteractive=0;
int optNoDefaultInclude=0;
string optIncludePath=DEF_INCLUDE_PATH;
ofstream *optLogfile=0;
int optBits=8*sizeof(int);
string optDumpPrefix="";
char optDumpFormat='a';
int optDebug=0;
int optAutoDump=-1;
int optLog=0;
int optLogState=0;
int optCheck=0;
int optTrace=0;
int optSyntax=0;
int optEcho=0;
int optTest=0;
int optShellEscape=1;
int optAllowRedefines=0;

#define OPTERR(s) { \
  err->report(errOPT,def);	\
  if(s) err->remark(s);		\
  return 1;			\
}

#define LOGVAL(v) { \
  if(!strchr("1yYtT0nNfF",arg[0])) OPTERR("Invalid boolean value");	\
  v=(strchr("1yYtT",arg[0])!=0);					\
}

int evalopt(char c,const char *arg,sDef *def,tError *err) {
  int n;
  char *p;
  switch(c) {
    case 's': n=strtol(arg,&p, 10);
  	      if(*p || n<0) OPTERR("Illegal seed value");
  	      srandom(n);
  	      break;
    case 'I': optIncludePath=arg;     		break;
    case 'p': optDumpPrefix=arg;   break;
    case 'f': if(!arg[0] || !strchr("hxdb",arg[0])) OPTERR("Invalid dump format");
  	      optDumpFormat=arg[0];
  	      if(optDumpFormat=='h') optDumpFormat='x';
  	      break;
    case 'd': LOGVAL(optDebug);	      		break;
    case 'a': LOGVAL(optAutoDump);		break;
    case 'l': LOGVAL(optLog);			break;
    case 'L': LOGVAL(optLogState);		break;
    case 'c': LOGVAL(optCheck);			break;
    case 'T': LOGVAL(optTrace);			break;
    case 'S': LOGVAL(optSyntax);		break;
    case 'E': LOGVAL(optEcho);			break;
    case 't': LOGVAL(optTest);	      		break;
    case 'e': LOGVAL(optShellEscape);		break;
    case 'r': LOGVAL(optAllowRedefines);   	break;
    case ':': OPTERR("Missing parameter");
    case '?':
    default:  OPTERR("Illegal option");
  };
  return 0;
}

#undef OPTERR

#define OPTERR(s) { \
  e.report(errOPT,0);	\
  if(s) e.remark(s);	\
  qclabort(e.errmsg());	\
}


void parseopt(int argc,char **argv) {
  int c=0,n;
  char *p;
  int ind;
  tError e;
  srandom(time(0));
  while(1) {
    c=getopt_long(argc,argv,QCL_SHORTOPTS,qcl_options,&ind);
    if(c==-1) break;
    switch(c) {
      case 'h':	printusage(argv[0]);
      		exit(0);
      case 'V': cerr << VERSION << ", " << COPYRIGHT << "\n" << DISCLAIMER << "\n";
      		exit(0);
      case 'i': optInteractive=1;       break;
      case 'n': optNoDefaultInclude=1;  break;
      case 'o': optLogfile=new ofstream(optarg);
                if(!optLogfile || !*optLogfile) {
		  optLogfile=0;
		  OPTERR(string("can't create logfile ")+optarg);
		};
		break;
      case 'b': n=strtol(optarg,&p, 10);
        	if(*p || n<2) OPTERR("Illegal number of qubits");
        	optBits=n;
        	break;
      default:  if(evalopt(c,optarg,0,&e)) qclabort(e.errmsg());
    }
  }
  if(argc==optind) optInteractive=1;
  if(optSyntax) optTest=1;
  if(optAutoDump==-1) optAutoDump=(optBits<=AUTODUMP_BITS ? 1 : 0);
}

#undef OPTERR
