/**********************************************************************

This file is part of the Quantum Computation Library (QCLIB).

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1996-1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/

#pragma implementation "bitvec.h"

#include "bitvec.h"

DEBUG( int nbitvecs=0; )

char* qc_check_txt     = "QCLIB: qc_check failed on condition ";
char* qc_check_stdmsg  = 
  "Can't continue. Provoking segfault to simplify backtracking.\n";
char* qc_check_msg  = qc_check_stdmsg;

char* qc_delete_txt    = "QCLIB: qc_delete warning";
char* qc_delarray_txt  = "QCLIB: qc_delarray warning";
char* qc_del_msg =
  "This should not happen. If the warning is reproducable, please\n"
  "send a bugreport to Bernhard Oemer <oemer@tph.tuwien.ac.at>.\n";

/* private members */

void bitvec::l_bitvec(word v) {
  int i,l;
  
  l=WORDS(len);
  pvec=new word[l];
  pvec[0]=v;
  for(i=1;i<l;i++) pvec[i]=0;
}

void bitvec::l_bitvec(const bitvec& v) {
  int i,l;

  l=WORDS(v.len);
  pvec=new word[l];
  for(i=0;i<l;i++) pvec[i]=v.pvec[i];
}

void bitvec::l_append(const bitvec& v) {
  int k;
  bitvec u(*this);
  setlen(len+v.len);
  *this=u+v;
}

bitvec bitvec::l_getbits(int i,int l) const {
  int k;
  word w;
  bitvec v(l);

  if(l<=BPW) { v.vec=l_getword(i,l); return v; };
  for(k=0;k<(l>>LDBPW);k++) v.pvec[k]=l_getword(i+(k<<LDBPW),BPW);
  if(l&DBITS) {
    k=l>>LDBPW;
    v.pvec[k]=l_getword(i+(k<<LDBPW),l&DBITS);
  };
  return v;
}

word bitvec::l_getword(int i,int l) const {
  int j,k;
  word w;
  
  j=i&DBITS;
  k=i>>LDBPW;
  w=(pvec[k]>>j)&MASK(l);
  if(l+j>BPW) w|=(pvec[k+1]<<(BPW-j))&MASK(l);
  return w;
}

void bitvec::l_setbits(int i,int l,word v) {
  int j,k;
  
  j=i&DBITS;
  k=i>>LDBPW;
  pvec[k]&=~(MASK(l)<<j);
  pvec[k]|=(MASK(l)&v)<<j;
  if(l+j<=BPW) return;
  pvec[k+1]&=~MASK(l+j-BPW);
  pvec[k+1]|=(MASK(l)&v)>>(BPW-j);
}
            
void bitvec::l_setbits(int i,const bitvec& v) {
  int j,k;

  for(k=0;k<(v.len>>LDBPW);k++) l_setbits(i+(k<<LDBPW),BPW,v.pvec[k]);
  if(v.len&DBITS) {
    k=v.len>>LDBPW;
    l_setbits(i+(k<<LDBPW),v.len&DBITS,v.pvec[k]);
  };
}

/* public members */

int bitvec::testless(const bitvec& v) const {
  int k;
  QC_CHECK(len==v.len);
  
  if(len<=BPW) return vec<v.vec;
  for(k=(len-1)>>LDBPW;k>=0;k--) {
    if(pvec[k]<v.pvec[k]) return 1;
    if(pvec[k]>v.pvec[k]) return 0;
  };
  return 0;
}

word bitvec::hashfunct() const {
  int k;
  word f=len;

  if(len<=BPW) return vec+(vec>>(BPW/2))+(vec>>(BPW/4))+(vec>>(BPW/8));
  for(k=0;k<WORDS(len);k++) f+=pvec[k]+(pvec[k]>>(BPW/2));
  return f;
}

bitvec& bitvec::not() {
  int k,l;

  if(len<=BPW) {
    vec=(~vec) & MASK(len);
  } else {
    l=WORDS(len);
    for(k=0;k<l;k++) pvec[k]=~pvec[k];
    if(len & DBITS) pvec[l-1]&=MASK(len & DBITS);
  };
  return *this;
}

/* other functions */

int bitvec_format=(BITVEC_FORMAT_HEX|BITVEC_FORMAT_PLAIN);
  
ostream& operator << (ostream& s,const bitvec& v) {
  int i,c;

  if(bitvec_format&BITVEC_FORMAT_KET) s << "|";
  if(bitvec_format&BITVEC_FORMAT_HEX) {
    s<<"0x";
    for(i=((v.length()-1) & ~3);i>=0;i-=4) {
      c=v.getword(i,(i+4<=v.length()) ? 4 : (v.length()-i));
      if(c<=9) s << (char)('0'+c); else s << (char)('a'+c-10);
    }
  } else {
    for(i=v.length()-1;i>=0;i--) s << (v[i] ? '1' : '0');
  }
  if(bitvec_format&BITVEC_FORMAT_KET) s << ">";
  return s;
}

istream& operator >> (istream& s,bitvec& v) {
  int j=0,i=0,h=0,k=0;
  char *b;
  char c;

  if(!(b=new char[v.length()+3])) goto error;  
  do {
    s >> c;
cerr << c;
    if(!s) goto error;
    if(c=='|') k=1;
  } while(c==' ' || c=='\t' || c=='|');
//  while((h && i<=(v.length()-1)/4) || (!h && i<v.length())) {
  while(1) {
    b[i++]=c;
    if(!(h && i<=(v.length()-1)/4) && !(!h && i<v.length())) break;
    s >> c;
cerr << c;
    if(!s) goto error;
    if(c=='x' && b[i-1]=='0') {
      h=1; s >> c; i=0;
      continue;
    }
    if(c>='A' && c<='F') c+='a'-'A';
    if((h && (c<'0' || c>'9') && (c<'a' || c>'f')) ||
       (!h && c!='0' && c!='1')) goto error;
  }
  b[i--]=0;
cerr << "i=" << i << "\n";
  if(k) {
    s >> c;
    if(c!='>') goto error;
  }
  v=0;
  if(h) {
    for(j=0;i>=0 && j<v.length();i--,j+=4)
      v.setbits(j,j+4<=v.length() ? 4 : v.length()-j,(word)
                (b[i]>='0' && b[i]<='9' ? b[i]-'0' : b[i]-'a'));
  } else {
    for(j=0;i>=0 && j<v.length();i--,j++)
      v.setbit(j,b[i]=='1');
  }
cerr << "read vec=" << v << "\n";
  return s;
  
error:

//fprintf(stderr,"err: [%s]\n",b);
cerr << "err: [" << b << "]\n";
  v.setlen(0);
  if(b) qc_delarray(b);
  return s;
}
