/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#pragma implementation

#include "quheap.h"
#include "extern.h"
#include "error.h"

#include <strstream.h>

int isStateModified=1;

term string2term(const string& s,int l) {
  term t(l);
  istrstream buf(s);
  buf >> t;
  return t;
}  
      
string term2string(const term& t,int binary) {
  ostrstream buf;
  buf << t;
  return string(buf.str());
}


class opFANOUT : public opGate {
  int inv;
public:
  opFANOUT(int n,int i) : opGate(2*n) { inv=i; }
  virtual term funct(const bitvec& v) const { 
    bitvec a(v.getbits(0,bits()/2));
    bitvec b(v.getbits(bits()/2,bits()/2));
    b^=a;
    a+=b;
    return term(a,1.0);
  };
  virtual opOperator *newclone() const { return new opFANOUT(bits()/2,inv); }
};


QuHeap::QuHeap(int n) {
  bs=new QuBaseState(n);
  ql=new qulist;
  nbits=n;
  mbits=bitvec(n).not();
  nfree=n;
  mfree=bitvec(n).not();
  ndirty=0;
  mdirty=bitvec(n);
  delbs=1;
  delay=0;
  parent=0;
}

QuHeap::QuHeap(quState *s,int d) {
  int n;
  bs=s;
  ql=new qulist;
  n=s->mapbits();
  nbits=n;
  mbits=bitvec(n).not();
  nfree=n;
  mfree=bitvec(n).not();
  ndirty=0;
  mdirty=bitvec(n);
  delbs=0;
  delay=d;
  parent=0;
}
 
QuHeap::QuHeap(QuHeap *h,int d) {
  bs=h->bs;
  ql=new qulist;
  nbits=h->nbits;
  mbits=h->mbits;
  nfree=h->nfree;
  mfree=h->mfree;
  ndirty=h->ndirty;
  mdirty=h->mdirty;
  delbs=0;
  delay=d;
  parent=h;
}

QuHeap::~QuHeap() {
  Pix p;
  if(bs && delbs) qcl_delete(bs);
  p=ql->first();
  while(p) {
    if((*ql)(p).s) delete (*ql)(p).s;
    ql->next(p);
  }
}

tValue QuHeap::qualloc(int n,BaseType t) {
  int i,o,l,ll; 
  quState *ss;
  if(!n || n>nfree) return tValue(tERROR);
  o=0;
  
  while(o+n<=nbits) {
    for(i=o;i<o+n;i++) {
      if(!mfree[i]) {
    	o=i+1;
    	break;
      };
    };
    if(i>=o) break;
  };
  if(o+n<=nbits) {
    ss=bs->newsubstring(n,o);
    nfree-=n;
    mfree&=~ ss->mapmask();
    return tValue(ss);
  };
  o=l=0;
  ss=0;
  while(o<nbits && l<n) {
    if(!mfree[o]) { o++; continue; }
    for(i=o+1;i<n || !mfree[i];i++);
    ll=i-o;
    if(l+ll>n) ll=n-l;
    if(ss) {
      ss=new quCombState(ss,bs->newsubstring(l,o));
    } else {
      ss=bs->newsubstring(l,o);
    };
    l+=ll;
  }
  if(!ss || l!=n || ss->mapbits()!=n) {
    qclerror(string("quheap: allocation failed, l=")+dec(l)+", o="+dec(o));
    if(ss) qcl_delete(ss);
    return tValue(tERROR);
  }
  nfree-=n;
  mfree&=~ ss->mapmask();
  return tValue(ss,t);    
}

int QuHeap::qufree(tValue var) {
  int n;
  bitvec m;
  quState *ss;
  if(!var.isQuExpr()) return 1;
  ss=var.qustate();
  n=ss->mapbits();
  m=ss->mapmask();
  if(n>nbits-nfree-ndirty ||
     !zero(m & mdirty) ||
     !zero(m & mfree)) return 1;
  nfree+=n;
  mfree|=m; 
  return 0;
}

tValue QuHeap::measure(tValue var) {
  bitvec m;
  if(!var.isQuExpr()) return tValue(tERROR);
  if(var.qustate()->mapbits()>8*sizeof(int)) 
    return tValue(tUNDEF);
  m=var.qustate()->measure();
  return(tValue((int)m.getword()));
}

int QuHeap::reset() {
  bs->reset();
  return 0;
}

int QuHeap::add(sRoutDef *d,SymTable *s,int inv) {
  quentry q;
  q.d=d; q.s=s; q.i=inv;
  ql->append(q);
  return 0;
}

int QuHeap::apply(SymTable *gl,tError *err,int inv) {
  Pix p;
  if(inv) {
    p=ql->last();
    while(p) {
      (*ql)(p).d->invoke((*ql)(p).s,gl,this,err,!(*ql)(p).i);
      ql->prev(p);
    }
  } else {
    p=ql->first();
    while(p) {
      (*ql)(p).d->invoke((*ql)(p).s,gl,this,err,(*ql)(p).i);
      ql->next(p);
    }
  }
  return 0;
}      

int QuHeap::addFanout(SymTable *gl,tError *err,tValue a,tValue b,int inv) {
  quentry q;
  void *c;

  if(!a.isQuExpr() || !b.isQuExpr) return 1;
  if(a.qustate()->mapbits() != b.qustate()->mapbits()) return 1;
  if(!(q.d=(sRoutDef*)gl->getDef(FANOUTID)) || !q.d->isRoutDef() || 
     !q.d->args()->first(c) || !q.d->args()->last(c)) return 1;
  q.s=new SymTab();
  q.s->put(q.d->args()->first(c),a);
  q.s->put(q.d->args()->last(c),b);
  q.i=inv;
  ql->append(q);
  return 0;
}

int QuHeap::applyFanout(SymTable *gl,tError *err,tValue a,tValue b,int inv) {
  sRoutDef *fo;
  SymTab sym;
  int r;
  void *c;

  if(!a.isQuExpr() || !b.isQuExpr) return 1;
  if(a.qustate()->mapbits() != b.qustate()->mapbits()) return 1;
  fo=(sRoutDef*)gl->getDef(FANOUTID);
  if( !fo || !fo->isRoutDef() ||
     !fo->args()->first(c) || !fo->args()->last(c)) return 1;
  sym.put(fo->args()->first(c),a);
  sym.put(fo->args()->last(c),b);
  r=fo->invoke(&sym,gl,this,err,inv);
  return r;
}      


int QuHeap::addSwap(SymTable *gl,tError *err,tValue a,tValue b,int inv) {
  quentry q;
  void *c;

  if(!a.isQuExpr() || !b.isQuExpr) return 1;
  if(a.qustate()->mapbits() != b.qustate()->mapbits()) return 1;
  if(!(q.d=(sRoutDef*)gl->getDef(SWAPID)) || !q.d->isRoutDef() || 
     !q.d->args()->first(c) || !q.d->args()->last(c)) return 1;
  q.s=new SymTab();
  q.s->put(q.d->args()->first(c),a);
  q.s->put(q.d->args()->last(c),b);
  q.i=inv;
  ql->append(q);
  return 0;
}

int QuHeap::applySwap(SymTable *gl,tError *err,tValue a,tValue b,int inv) {
  sRoutDef *sw;
  SymTab sym;
  int r;
  void *c;

  if(!a.isQuExpr() || !b.isQuExpr) return 1;
  if(a.qustate()->mapbits() != b.qustate()->mapbits()) return 1;
  sw=(sRoutDef*)gl->getDef(SWAPID);
  if( !sw || !sw->isRoutDef() ||
     !sw->args()->first(c) || !sw->args()->last(c)) return 1;
  sym.put(sw->args()->first(c),a);
  sym.put(sw->args()->last(c),b);
  r=sw->invoke(&sym,gl,this,err,inv);
  return r;
}      

string QuHeap::prtstr() {
  string ostr;
  Pix p;
  if(parent) ostr+=parent->prtstr();
  ostr+="{ ";
  p=ql->first();
  while(p) {
    if((*ql)(p).i) ostr+="!";
    ostr+=(*ql)(p).d->id();
    ostr+=(*ql)(p).s->prtstr();
    ostr+=" ";
    ql->next(p);
  }
  ostr+="}";
  return ostr;
}

int QuHeap::load(ifstream& f) {
  int i,l=-1;
  term t;
  string s;
  char buf[1024];

  term_format=TERM_FORMAT_LIST;
  f.scan("QCL begin state %d qubits\n",&l);
  cerr << l << "=l\n";
  if(l!=nbits) return 2;
  bs->opbegin();
  while(!f.eof()) {
    f.getline(buf,1024);
    s=buf;
    if(s=="QCL end state") break;
    t=string2term(s,l);
    if(t.nbits()==0) {
      return 1;
    }
    bs->opadd(t);
  }
  bs->opend();
  return 0;
}

int QuHeap::save(ofstream& f) {
  int i;
  string s;
  
  term_format=TERM_FORMAT_LIST;
  f << "QCL begin state " << bs->basebits() << " qubits\n";
  if(!f) return 1;
  for(i=0;i<bs->baseterms();i++) {
    f << bs->baseterm(i) << "\n";
    if(!f) return 1;
  }
  f << "QCL end state\n";
  if(!f) return 1;
  return 0;
}
