/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#pragma implementation

#include <stdlib.h>
#include <stdio.h>

#include "syntax.h"
#include "error.h"
#include "quheap.h"
#include "debug.h"
#include "options.h"

#define ERR 1
#define RET 2

#define EXERR(e,s) { 		\
  err->report(e,this);			\
  if(s) err->remark(s);			\
  err->line(prtstr(),0);		\
  if(optDebug) qclshell(loc,gl,qh,this,err,DB_ERROR); \
  return ERR;				\
}

int sRoutDef::invoke(SymTable *loc,SymTable *gl,QuHeap *qh,tError *err,int inv) {
  SymTab l;
  SymTabComb lcomb((SymTable*)&l,loc);
  int delay=(inv || isBennet());
  QuHeap lqh(qh,delay);
  int r,i;
  void *c;
  sDef *p;
  tValue *pv,*qv;
  tValue v;
  sRoutDef *fo;
  SymTab *psym;
  int f;
  ostream *log=&cout;
  
  if(optCheck) {
    for_plist(p,args(),c) {
      if(!inv && p->type().isQuVoid() || p->type().isQuScr()) {
        pv=loc->getRef(p->id());
	if(!pv || !pv->isQuExpr()) 
	  EXERR(errINT,"invalid quantum parameter");
	i=pv->qustate()->mapbits();
	if(pv->qustate()->prob(bitvec(i))<(1.0-EPSILON_CHECK))
	  EXERR(errMEM,"void or scratch register not empty");
      }
    }
  }
  
  if(isExtern()) {
    if(optLog) {
      if(optLogfile) log=optLogfile;
      *log << "@ ";
      if(inv) *log << "!";
      *log << id() << "(";
      f=0;
      for_plist(p,args(),c) {
        if(f) *log << ","; else f=1;
        *log << p->prtstr() << "=";
	if(pv=loc->getRef(p->id())) {
	  *log << pv->str();
	} else {
	  *log << "undefined";
	}
      }
      *log << ")\n";
    };
    if(!optTest) {
      isStateModified=1;
      r=(pext)(loc,gl,qh,err,inv);
      if(optLogState) printbase(*log,*qh->state(),"% ","\n");
      return r;
    } else {
      return 0;
    };
  }

  if(isBennet()) {
    for_plist(p,args(),c) {
      if(p->type().isQuVoid()) {
        pv=loc->getRef(p->id());
        if(!pv || !pv->isQuExpr()) 
	  EXERR(errINT,"invalid quantum parameter");
        v=lqh.qualloc(pv->qustate()->mapbits());
	if(v.isError())
	  EXERR(errMEM,"can't allocate internal scratch space");
	l.put(p,v);
      }
    }
  }
  if(pdefs->define(&lcomb,gl,&lqh,this,err)) return ERR;
  r=body()->exec(&lcomb,gl,&lqh,this,err);
  if(r) return r;
  if(delay) {
    if(isBennet()) {
      if(lqh.apply(gl,err,0)) return ERR;
      for_plist(p,args(),c) {
      	if(p->type().isQuVoid()) {
      	  pv=loc->getRef(p->id());
      	  if(!pv || !pv->isQuExpr()) 
	    EXERR(errINT,"parameter "+p->id()+" not found");
      	  qv=l.getRef(p->id());
      	  if(!qv || !qv->isQuExpr()) 
	    EXERR(errINT,"temporary register not found");
      	  if(lqh.applyFanout(gl,err,*qv,*pv,inv)) return ERR;
      	}
      }
      if(lqh.apply(gl,err,1)) return ERR;
    } else {
      if(lqh.apply(gl,err,inv)) return ERR;
    }
  }
  
  if(optCheck) {
    term t;
    for_plist(p,args(),c) {
      if(inv && p->type().isQuVoid() || p->type().isQuScr()) {
        pv=loc->getRef(p->id());
	if(!pv || !pv->isQuExpr()) return ERR;
	i=pv->qustate()->mapbits();
	if(pv->qustate()->prob(bitvec(i))<(1.0-EPSILON_CHECK))
	  EXERR(errMEM,"void or scratch register not empty");
      }
    }
    for(i=0;i<qh->state()->baseterms();i++) {
      t=qh->state()->baseterm(i);
      if(!zero(t.vect()&qh->mFree()) && norm(t.ampl())>EPSILON_CHECK)
        EXERR(errMEM,"quantum heap is corrupted");
    }
  }
  return 0;
}

int sFunctDef::invoke(SymTable *loc,SymTable *gl,QuHeap *qh,tError *err,int inv) {
  SymTab l;
  SymTabComb lcomb((SymTable*)&l,loc);
  int r;
  if(pdefs->define(&lcomb,gl,qh,this,err)) return ERR;
  r=body()->exec(&lcomb,gl,qh,this,err);
  return r!=RET;
}


#define ENTRY(s,n) case n: return s;

string sObject::objstr() const {
  switch(object()) {
    ENTRY("sVOID",sVOID);

    ENTRY("sOBJECT",sOBJECT);
    ENTRY("sLIST",sLIST);

    ENTRY("sEXPR",sEXPR);
    ENTRY("sEXPRLIST",sEXPRLIST);
    ENTRY("sCONST",sCONST);
    ENTRY("sVAR",sVAR);
    ENTRY("sSUBSCRIPT",sSUBSCRIPT);
    ENTRY("sSUBRANGE",sSUBRANGE);
    ENTRY("sFUNCTCALL",sFUNCTCALL);
    ENTRY("sUNOP",sUNOP);
    ENTRY("sNEG",sNEG);
    ENTRY("sNOT",sNOT);
    ENTRY("sLENGTH",sLENGTH);
    ENTRY("sBINOP",sBINOP);
    ENTRY("sADD",sADD);
    ENTRY("sSUB",sSUB);
    ENTRY("sMULT",sMULT);
    ENTRY("sDIV",sDIV);
    ENTRY("sMOD",sMOD);
    ENTRY("sPOW",sPOW);
    ENTRY("sLESS",sLESS);
    ENTRY("sEQUAL",sEQUAL);
    ENTRY("sLEEQ",sLEEQ);
    ENTRY("sNOTEQ",sNOTEQ);
    ENTRY("sAND",sAND);
    ENTRY("sOR",sOR);
    ENTRY("sXOR",sXOR);
    ENTRY("sCONCAT",sCONCAT);
    ENTRY("sBASEFUNCT",sBASEFUNCT);
    ENTRY("sEXP",sEXP);
    ENTRY("sSIN",sSIN);
    ENTRY("sCOS",sCOS);
    ENTRY("sTAN",sTAN);
    ENTRY("sCOT",sCOT);
    ENTRY("sSINH",sSINH);
    ENTRY("sCOSH",sCOSH);
    ENTRY("sTANH",sTANH);
    ENTRY("sCOTH",sCOTH);
    ENTRY("sABS",sABS);
    ENTRY("sRE",sRE);
    ENTRY("sIM",sIM);
    ENTRY("sCONJ",sCONJ);
    ENTRY("sFLOOR",sFLOOR);
    ENTRY("sCEIL",sCEIL);
    ENTRY("sSQRT",sSQRT);
    ENTRY("sLISTFUNCT",sLISTFUNCT);
    ENTRY("sLOG",sLOG);
    ENTRY("sRANDOM",sRANDOM);
    ENTRY("sMIN",sMIN);
    ENTRY("sMAX",sMAX);
    ENTRY("sGCD",sGCD);
    ENTRY("sLCM",sLCM);

    ENTRY("sSTMT",sSTMT);
    ENTRY("sSTMTLIST",sSTMTLIST);
    ENTRY("sCALL",sCALL);
    ENTRY("sASSIGN",sASSIGN);
    ENTRY("sTRANS",sTRANS);
    ENTRY("sFANOUT",sFANOUT);
    ENTRY("sSWAP",sSWAP);
    ENTRY("sFOR",sFOR);
    ENTRY("sIF",sIF);
    ENTRY("sLOOP",sLOOP);
    ENTRY("sWHILE",sWHILE);
    ENTRY("sUNTIL",sUNTIL);
    ENTRY("sRETURN",sRETURN);
    ENTRY("sINPUT",sINPUT);
    ENTRY("sPRINT",sPRINT);
    ENTRY("sEXIT",sEXIT);
    ENTRY("sMEASURE",sMEASURE);
    ENTRY("sINCLUDE",sINCLUDE);
    ENTRY("sRESET",sRESET);
    ENTRY("sDUMP",sDUMP);
    ENTRY("sINSPECT",sINSPECT);
    ENTRY("sLOAD",sLOAD);
    ENTRY("sSAVE",sSAVE);
    ENTRY("sSET",sSET);
    ENTRY("sSHELL",sSHELL);

    ENTRY("sDEF",sDEF);
    ENTRY("sDEFLIST",sDEFLIST);
    ENTRY("sROUTDEF",sROUTDEF);
    ENTRY("sFUNCTDEF",sFUNCTDEF);
    ENTRY("sPROCDEF",sPROCDEF);
    ENTRY("sQUOPDEF",sQUOPDEF);
    ENTRY("sQUFUNDEF",sQUFUNDEF);
    ENTRY("sARGDEF",sARGDEF);
    ENTRY("sVARDEF",sVARDEF);
    ENTRY("sCONSTDEF",sCONSTDEF);

    ENTRY("sEOF",sEOF);
    ENTRY("sMSG",sMSG);
    default: ;
  }
  return "UNKNOWN_OBJECT"; 
}

#undef ENTRY
