/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#pragma implementation

#include "syntax.h"
#include "error.h"
#include "options.h"

#include <stdlib.h>

#define TRACE() if (optTrace) { cerr << "TYPECHECK"; err->trace(this,loc,gl,0,def); }


#define TYPERR(e,d,s) { 		\
  err->report(e,d);			\
  if(s) err->remark(s);			\
  if(1 || isStmt() || isDef()) 		\
    err->line(prtstr(),0);		\
  return 1;				\
}

#define CHECK(v,d) {			\
  if(v && v->typecheck(loc,gl,d,err)) { \
    if(isStmt() && v->isExpr())	{	\
      err->line(prtstr(),0);		\
    };					\
    return 1;				\
  }					\
}


/***********************************************/
/***************** Definitions *****************/
/***********************************************/


int sVarDef::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(plen,def);
  CHECK(pinit,def);
  if(def) {
    if(loc && loc->getDef(id())) 
      TYPERR(errSCOPE,def,"Local symbol "+id()+" already defined");
  } else {
    if(gl && gl->getDef(id())) 
      TYPERR(errSCOPE,def,"Global symbol "+id()+" already defined");
  };
  if(type().isQuExpr()) {
    if(def && def->isFunctDef()) 
      TYPERR(errSCOPE,def,"Quantum Variables may non be defined within functions");
    if(type()==tQUSCR && (!def || !def->isQuFunDef()))
      TYPERR(errSCOPE,def,"Scratch Space may only be allocated within quantum functions");
    if(type()==tQUVOID)
      TYPERR(errSCOPE,def,"Void Registers have to be arguments");
    if(plen && pinit) 
      TYPERR(errDEF,def,"Quantum variable can either be allocated or defined as reference, not both");
    if(plen && !plen->type().isInt()) 
      TYPERR(errDEF,def,"Invalid length");
    if(pinit && !pinit->type().isQuExpr()) 
      TYPERR(errTYPMIS,def,"Invalid initialiser for quantum register reference");
    if(!plen && !pinit) 
      TYPERR(errDEF,def,"Quantum variable must be allocated or initialized");
  } else {
    if(plen) TYPERR(errDEF,def,"Length cannot be specified with this type");
    if(pinit && pinit->type()!=type()) 
      TYPERR(errTYPMIS,def,"Invalid initialisation");
  };
  if(def) if(loc->put(this)) 
    TYPERR(errINT,def,"can't add symbol to symtab");
  if(def && type()==tQUSCR) ((sQuFunDef*)def)->setBennet(1);
  return 0;
}

int sConstDef::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);

  if(def) {
    if(loc && loc->getDef(id())) 
      TYPERR(errSCOPE,def,"Local symbol "+id()+" already defined");
  } else {
    if(gl && gl->getDef(id())) 
      TYPERR(errSCOPE,def,"Global symbol "+id()+" already defined");
  };
  if(pexpr->type().isQuExpr()) 
    TYPERR(errTYPMIS,def,"Quantum variables cannot be defined as constant");
  settype(pexpr->type());
  if(def) if(loc->put(this)) TYPERR(errINT,def,"can't add symbol "+id()+" to symtab");
  return 0;
}

int sRoutDef::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  SymTab l,g;
  void *c;
  int i,r;
  int f=0;
  TRACE();
  
  if(gl && gl->getDef(id())) 
    TYPERR(errSCOPE,def,"Global symbol "+id()+" already defined");
  if(def) 
    TYPERR(errSCOPE,def,"subroutines may only be defined in global scope");
  CHECK(pargs,def);
  for_plist(p,pargs,c) {
    if(l.getRef(p->id())) TYPERR(errINVPAR,def,"duplicate parameter");
    if(isFunctDef() && p->type().isQuExpr()) 
      TYPERR(errINVPAR,def,"Functions may not depend on quantum parameters");
    if(p->type().isQuReg()) f=1;  
    l.put(p);
  };
  if(isExtern()) {
    pext=0;
    for(i=0;ExtRoutTable[i].rout;i++) {
      if(id()==string(ExtRoutTable[i].id)) {
        pext=ExtRoutTable[i].rout;
        break;
      }
    }
    if(!pext) TYPERR(errEXT,def,"external routine "+id()+" not found");
  } else {
    if(isFunctDef()) g.put(this);
    if(pdefs->typecheck(&l,&SymTabComb(&g,gl),this,err)) return 1;
    if(isQuFunDef() && f && (sQuFunDef*)this->isBennet())
      TYPERR(errINVTYP,def,"local scratch registers can't be used with qureg arguments");  
    if(pbody->typecheck(&l,&SymTabComb(&g,gl),this,err)) return 1;
  }
  return 0;
}

int sDefList::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  void *c;
  TRACE();
  for_plist(p,this,c) if(p->typecheck(loc,gl,def,err)) return 1;
  return 0;
}


/***********************************************/
/***************** Statements ******************/
/***********************************************/


int sStmtList::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sStmt *p;
  void *c;
  TRACE();
  for_plist(p,this,c) if(p->typecheck(loc,gl,def,err)) return 1;
  return 0;
}

int sCall::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sRoutDef *d;
  sExpr *p;
  sDef *q;
  string s;
  void *cp,*cq;
  tType pt,qt;
  TRACE();
  d=(sRoutDef*)gl->getDef(id());
  if(!d || !d->isRoutDef() || d->isFunctDef()) 
    TYPERR(errSYMBOL,def,"Undefined procedure or operator "+id());
  if(def && !def->isRoutDef()) 
    TYPERR(errINT,def,"Local call within non subroutine definition");
  if(def && def->isFunctDef()) 
    TYPERR(errSCOPE,def,"Calls not allowed within functions");
  if(def && def->isQuDef() && d->isProcDef())
     TYPERR(errSCOPE,def,"Procedure call within operator");
  if(def && def->isQuFunDef() && d->isQuOpDef())
     TYPERR(errSCOPE,def,"operator called within qufunct");
  for_pplist(p,ppar,cp,q,d->args(),cq,p || q) {
    if(!p || !q) TYPERR(errPARMIS,def,"unmatching argument number");
    CHECK(p,def);
    pt=p->type(); qt=q->type();
    if(pt!=qt && !(pt.isQuExpr() && qt.isQuExpr()))
      TYPERR(errPARMIS,def,"unmatching argument types");
    if(pt.isQuConst() && !qt.isQuConst())
      TYPERR(errPARMIS,def,"quconst used as non-const argument to "+id()); 
    p->settype(q->type());
  };
  return 0;
}

int sAssign::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  TRACE();
  CHECK(pvar,def);
  CHECK(pexpr,def);
  p=loc->getDef(pvar->id());
  if(!p && def && def->object()!=sPROCDEF) 
    TYPERR(errSYMBOL,def,"Unknown local variable "+pvar->id());
  if(!p) p=gl->getDef(pvar->id());
  if(!p || p->object()!=sVARDEF) 
    TYPERR(errSYMBOL,def,"Unknown variable "+pvar->id());
  if(p->type().isQuExpr()) 
    TYPERR(errINVTYP,def,"assignment to quantum variable");
  if(p->type()!=pexpr->type()) 
    TYPERR(errTYPMIS,def,"invalid assignment");
  return 0;
}

int sTrans::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  string opid= object()==sFANOUT ? FANOUTID : SWAPID;
  TRACE();
  CHECK(plval,def);
  CHECK(prval,def);
  
  if(!gl->getDef(opid)) 
    TYPERR(errINT,def,"no "+opid+" operator defined");
  if(!plval->type().isQuExpr() || !prval->type().isQuExpr()) 
    TYPERR(errINVTYP,def,opid+" is only defined on quantum expressions");
    if(prval->type().isQuConst() || 
       (object()==sSWAP && plval->type().isQuConst()))
      TYPERR(errINVTYP,def,"constant target register for "+opid);
  return 0;
}

int sFor::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  sConstDef *pv;
  SymTab sym;
  int r;
  TRACE();
  CHECK(pvar,def);
  p=loc->getDef(pvar->id());
  if(!p && def && def->object()!=sPROCDEF) 
    TYPERR(errSYMBOL,def,"Unknown local variable "+pvar->id());
  if(!p) p=gl->getDef(pvar->id());
  if(!p || p->object()!=sVARDEF) 
    TYPERR(errSYMBOL,def,"Invalid Symbol type");
  CHECK(pfrom,def);
  CHECK(pto,def);
  if(pstep) CHECK(pstep,def);
  if(!pfrom->type().isInt() || !pto->type().isInt() ||
    (pstep && !pstep->type().isInt())) 
    TYPERR(errINVTYP,def,"for loop parameters must be integer");
  pv=new sConstDef(new tId(p->id()),(sExpr*)new sConst(tValue(0)));
  if(!pv) TYPERR(errINT,def,"alloc failed");
  pv->settype(tType(tINTEGER));
  sym.put(pv);
  r=pbody->typecheck(&SymTabComb(&sym,loc),gl,def,err);
  qcl_delete(pv);
  return r;
}

int sIf::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pcond,def);
  if(!pcond->type().isBoolean()) 
    TYPERR(errTYPMIS,def,"not boolean if condition");
  CHECK(pthen,def);
  CHECK(pelse,def);
  return 0;
}

int sLoop::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pcond,def);
  if(!pcond->type().isBoolean()) 
    TYPERR(errTYPMIS,def,"not boolean loop condition");
  CHECK(pbody,def);
  return 0;
}

int sReturn::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);
  if(!def || !def->isFunctDef()) 
    TYPERR(errSCOPE,def,"return statement outside function");
  if(pexpr->type()!=def->type()) 
    TYPERR(errTYPMIS,def,"return expression doesn't match function type");
  return 0;
}

int sInput::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  TRACE();
  CHECK(pprompt,def);
  CHECK(pvar,def);
  if(def && !def->isProcDef()) 
    TYPERR(errSCOPE,def,"input not allowed in this scope");
  if(pprompt && !pprompt->type().isString())
    TYPERR(errINVTYP,def,"input prompt is no string");

  p=loc->getDef(pvar->id());
  if(!p && def && def->object()!=sPROCDEF) 
    TYPERR(errSYMBOL,def,"Unknown local variable "+pvar->id());
  if(!p) p=gl->getDef(pvar->id());
  if(!p || p->object()!=sVARDEF) 
    TYPERR(errSYMBOL,def,"Unknown variable "+pvar->id());
  if(pvar->type().isQuExpr())
    TYPERR(errINVTYP,def,"cannot input quantum variables");
  return 0;
}

int sPrint::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(plist,def);
  return 0;
}

int sExit::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pmsg,def);
  if(pmsg && !pmsg->type().isString()) 
    TYPERR(errSCOPE,def,"error message is not a string");
  return 0;
}

int sMeasure::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p;
  TRACE();
  CHECK(pstate,def);
  CHECK(pvar,def);
  if(def && !def->isProcDef()) 
    TYPERR(errSCOPE,def,"measurement is not allowed in this scope");
  if(!pstate || !pstate->type().isQuExpr())
    TYPERR(errINVTYP,def,"quantum state required");

  if(pvar) {
    p=loc->getDef(pvar->id());
    if(!p && def && def->object()!=sPROCDEF) 
      TYPERR(errSYMBOL,def,"Unknown local variable "+pvar->id());
    if(!p) p=gl->getDef(pvar->id());
    if(!p || p->object()!=sVARDEF) 
      TYPERR(errSYMBOL,def,"Unknown variable "+pvar->id());
    if(!pvar->type().isInt())
      TYPERR(errINVTYP,def,"quantum eigenstates are integers");
  }
  return 0;
}

int sInclude::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  TYPERR(errINT,def,"uncaught include");
  return 1;
}


int sReset::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  if(def && !def->isProcDef()) 
    TYPERR(errSCOPE,def,"Reset of quantum state is not allowed in this  scope");
  return 0;
}

/***********************************************/
/****************** Commands *******************/
/***********************************************/


int sDump::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);
  if(pexpr && !pexpr->type().isQuExpr())
    TYPERR(errINVTYP,def,"Quantum expression required");
  return 0;
}

int sInspect::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sExpr *p;
  void *c;
  TRACE();
  if(plist) {
    for_plist(p,plist,c) {
      if(!p || p->object()!=sVAR)
  	TYPERR(errINVTYP,def,"Can only list symbols");
    };
  };
  return 0;
}

int sLoad::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);
  if(pexpr && !pexpr->type().isString())
    TYPERR(errINVTYP,def,"Filename must be a string");
  return 0;
}

int sSave::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);
  if(pexpr && !pexpr->type().isString())
    TYPERR(errINVTYP,def,"Filename must be a string");
  return 0;
}

int sSet::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  int i;
  tValue v;
  string s;
  TRACE();
  CHECK(pexpr,def);
  if(pexpr && pexpr->type().isQuExpr())
    TYPERR(errINVTYP,def,"quantum expression as option argument"); 
  for(i=0;qcl_options[i].val;i++) {
    if(opt==string(qcl_options[i].name)) goto cont;
  }
  TYPERR(errOPT,def,"illegal option "+opt);
cont:
  if(qcl_options[i].has_arg && !pexpr)
    TYPERR(errOPT,def,"missing argument for option "+opt);
  if(!qcl_options[i].has_arg && pexpr)
    TYPERR(errOPT,def,"option "+opt+" takes no argument");
  return 0;
}

int sShell::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  return 0;
}

/***********************************************/
/***************** Expressions *****************/
/***********************************************/


int sExprList::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sExpr *p;
  void *c;
  TRACE();
  for_plist(p,this,c) {
    CHECK(p,def);
  };
  return 0;
}

int sConst::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  if(!_val.type().isDefined) TYPERR(errINT,def,
    "undefined constant");
  settype(_val.type());
  return 0;
}

int sVar::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sDef *p,*q=0;
  TRACE();
  p=loc->getDef(id());
  if(!p) q=gl->getDef(id());
  if(!p && q && q->isConstDef()) p=q;
  if(!p && def && def->object()!=sPROCDEF) 
    TYPERR(errSYMBOL,def,"Unknown local variable "+id());
  if(!p) p=q;
  if(!p || !p->isValueDef()) 
    TYPERR(errSYMBOL,def,"Unknown variable or constant "+id());
  if(p->type().isQuExpr() && def && def->isFunctDef())
    TYPERR(errINVTYP,def,"quantum variables may not be accessed within functions");
  settype(p->type());
  return 0;
}

int sSubscript::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pvar,def);
  CHECK(psub,def);
  if(!pvar->type().isQuExpr()) 
    TYPERR(errINVTYP,def,"bit selection not on quantum variable");
  if(!psub->type().isInt()) 
    TYPERR(errINVTYP,def,"subscript not integer");
  settype(pvar->type());
  return 0;
}

int sSubRange::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pvar,def);
  CHECK(pexpr1,def);
  CHECK(pexpr2,def);
  if(!pvar->type().isQuExpr()) 
    TYPERR(errINVTYP,def,"bit selection on on quantum variable");
  if(!pexpr1->type().isInt() || !pexpr2->type().isInt()) 
    TYPERR(errINVTYP,def,"subrange parameters not integer");
  settype(pvar->type());
  return 0;
}

int sUnOp::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  TRACE();
  CHECK(pexpr,def);
  switch(object()) {
    case sNEG:
      if(!pexpr->type().isNumber()) 
        TYPERR(errINVTYP,def,"Unary Minus: argument must be number");
      settype(pexpr->type());
      break;
    case sNOT:
      if(!pexpr->type()==tBOOLEAN) 
        TYPERR(errINVTYP,def,"Negation: argument must be boolean");
      settype(pexpr->type());
      break;
    case sLENGTH:
      if(!pexpr->type().isQuExpr()) 
        TYPERR(errINVTYP,def,"Length operator is only defined for quantum expressions");
      settype(tType(tINTEGER));
      break;
    default:
      TYPERR(errINT,def,"Invalid unary operator");
  }
  return 0;
}

int sBinOp::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  tType t1,t2;
  TRACE();
  CHECK(pexpr1,def);
  CHECK(pexpr2,def);
  t1=pexpr1->type();
  t2=pexpr2->type();
  switch(object()) {
    case sMOD:
      if(t1!=tINTEGER || t2!=tINTEGER) 
        TYPERR(errINVTYP,def,"Modulus arguments must be integer");    
    case sADD:
    case sSUB:
    case sMULT:
    case sDIV:
    case sPOW:    
      if(!t1.isNumber() || !t2.isNumber()) 
        TYPERR(errINVTYP,def,"Non numeric argument to binary arithmetic operator");
      if(t1==tINTEGER && t2==tINTEGER) { 
        settype(tType(tINTEGER));
        break;
      }
      if(t1==tCOMPLEX || t2==tCOMPLEX) {
        settype(tType(tCOMPLEX));
        break;
      }
      settype(tType(tREAL));
      break;
      
    case sLEEQ:
    case sLESS:
      if(!t1.isOrdered() || !t2.isOrdered())
        TYPERR(errINVTYP,def,"comparison operator with unordered type");
    case sNOTEQ:
    case sEQUAL:
      if(t1.isQuExpr() || t2.isQuExpr()) 
        TYPERR(errINVTYP,def,"can't compare quantum expressions");
      if(t1.isBoolean() || t2.isBoolean())
        TYPERR(errINVTYP,def,"can't compare boolean values"); 
      if((!t1.isNumber() || !t2.isNumber()) && t1!=t2)
        TYPERR(errINVTYP,def,"comparison type mismatch");
      settype(tType(tBOOLEAN));
      break;  
    
    case sAND:
    case sOR:
    case sXOR:
      if(t1!=tBOOLEAN || t2!=tBOOLEAN) 
        TYPERR(errINVTYP,def,"logical operator with non boolean arguments");
      settype(tType(tBOOLEAN));
      break;  

    case sCONCAT:
      if(t1.isQuExpr() && t2.isQuExpr()) {
        if(t1==t2) { settype(t1); break; };
        if(t1.isQuConst() || t2.isQuConst()) { settype(tType(tQUCONST)); break; }
        settype(tType(tQUREG));
        break;
      }
      if(t1.isString() && t2.isString()) {
        settype(tType(tSTRING));
        break;
      }
      TYPERR(errINVTYP,def,"concatenation of invalid types");

    default:
      TYPERR(errINT,def,"Invalid binary operator");
  }
  return 0;
}

int sFunctCall::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  sRoutDef *d;
  sExpr *p;
  sDef *q;
  string s;
  void *cp,*cq;
  TRACE();
  d=(sRoutDef*)gl->getDef(id());
  if(!d || !d->isFunctDef()) 
    TYPERR(errSYMBOL,def,"Undefined function "+id());
  if(def && !def->isRoutDef()) 
    TYPERR(errINT,def,"Local call within non subroutine definition");
  for_pplist(p,ppar,cp,q,d->args(),cq,p || q) {
    if(!p || !q) TYPERR(errPARMIS,def,"unmatching argument number");
    if(p->typecheck(loc,gl,def,err) || p->type()!=q->type())
      TYPERR(errPARMIS,def,"unmatching argument types");
  };
  settype(d->type());
  return 0;
}

int sBaseFunct::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  tValue v;
  tComplex z;
  tReal r;
  tType t;
  TRACE();
  CHECK(pexpr,def);
  t=pexpr->type();
  if(isTrigFunct() || object()==sEXP || object()==sSQRT) {
    if(!t.isNumber()) 
      TYPERR(errTYPMIS,def,"argument is not a number");
    if(pexpr->type().isComplex()) 
      settype(tType(tCOMPLEX));
    else
      settype(tType(tREAL));    
  } else {
    if(!t.isNumber()) 
      TYPERR(errTYPMIS,def,"argument is not a number");
    switch(object()) {
      case sABS	 :
        if(t.isInt()) 
          settype(tType(tINTEGER));
        else
          settype(tType(tREAL)); 
        break;   
      case sRE	 : 
      case sIM	 :
        settype(tType(tREAL));
        break;
      case sCONJ   :
        settype(tType(tCOMPLEX));
        break;
      case sFLOOR  : 
      case sCEIL   : 
        if(t.isComplex())
          TYPERR(errTYPMIS,def,"ordered type required");
        settype(tType(tINTEGER));
        break;
      default:
        TYPERR(errINT,def,"unknown base function");
    }
  }
  return 0;
}

int sListFunct::typecheck(SymTable *loc,SymTable *gl,sDef *def,tError *err) {
  tValue v;
  tComplex z;
  tReal r;
  tType t;
  int argc;
  void *c;
  sExpr *p;
  TRACE();
  CHECK(pexpr,def);
  if(pexpr1) t=pexpr1->type();
  argc=pexpr->n();
  switch(object()) {
    case sLOG:
      if(!pexpr1 || argc>2)
    	TYPERR(errSYNTAX,def,"invalid number of argument to log");
      if(!pexpr1->type().isNumber() || (pexpr2 && !pexpr2->type().isOrdered()))
    	TYPERR(errTYPMIS,def,"invalid log argument");
      if(t.isComplex())
    	settype(tType(tCOMPLEX));
      else
    	settype(tType(tREAL));
      return 0;
    case sRANDOM:
      if(def && !def->isProcDef())
        TYPERR(errSCOPE,def,"function random is not allowed in this scope"); 
      if(argc) TYPERR(errSYNTAX,def,"function random takes no arguments");
      settype(tType(tREAL));
      return 0;
    case sMIN:
    case sMAX:
      if(!argc) TYPERR(errSYNTAX,def,"missing argument");
      settype(tType(tINTEGER));
      for_plist(p,pexpr,c) {
        if(!p->type().isOrdered()) 
          TYPERR(errTYPMIS,def,"argument of unordered type");
        if(p->type().isReal()) settype(tType(tREAL));
      }
      return 0;
    case sGCD:
    case sLCM:
      if(!argc) TYPERR(errSYNTAX,def,"missing argument");
      for_plist(p,pexpr,c) {
        if(!p->type().isInt()) 
          TYPERR(errTYPMIS,def,"integer required");
      }
      settype(tType(tINTEGER));
      return 0;
  }
  TYPERR(errINT,def,"unknown list function");
}    
