/**********************************************************************

This file is part of the Quantum Computation Language QCL.

(c) Copyright by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 1998

This program comes without any warranty; without even the implied 
warranty of merchantability or fitness for any particular purpose.

     This program is free software under the terms of the 
     GNU General Public Licence (GPL) version 2 or higher

************************************************************************/


#ifndef TYPES_H
#define TYPES_H 1

#pragma interface

#include "qc.h"

#include <String.h>
#include <complex.h>
#include <ostream.h>

#include <stdio.h>

#ifdef QCL_DEBUG
#define qcl_delete(p) { \
  if(!p) { \
    cerr << "qcl_delete warning in " << __FILE__ << ":" << __LINE__<<  "\n"; \
  } else { delete p; p=0; } \
}
#define qcl_delarray(p) { \
  if(!p) { \
    cerr << "qcl_delarray warning in " << __FILE__ << ":" << __LINE__ << "\n"; \
  } else { delete[] p; p=0; } \
}
#else
#define qcl_delete(p) { if(p) { delete p; p=0; } }
#define qcl_delarray(p) { if(p) { delete[] p; p=0; } }
#endif

typedef double tReal;

#ifdef __STD_COMPLEX
typedef double_complex tComplex;
#else
typedef complex tComplex;
#endif

#ifndef __BASTRING__
typedef String string;
#endif

#define NOINDENT (-1)
#define INDENT 0

extern const char *FALSE_STR;
extern const char *TRUE_STR;
extern const char *BASETYPE_STR[];

typedef string tId;

enum boolean { FALSE = 0, TRUE = 1 };

enum BaseType { tUNDEF, tERROR, tVOID, tQUOP, tQUFUN,
		tBOOLEAN, tINTEGER, tREAL, tCOMPLEX, tSTRING,
                tQUCONST, tQUREG, tQUVOID, tQUSCR };
                
// Class Hierarchy

enum ObjType { 
sVOID			= 0,

sOBJECT			= 1000, 
  sLIST			= 1010, 

  sEXPR			= 2000, 
  sEXPRLIST		= 2010,
    sCONST		= 2020, 
    sVAR		= 2030, 
    sSUBSCRIPT		= 2040, 
    sSUBRANGE		= 2050, 
    sFUNCTCALL		= 2060, 
    sUNOP		= 2100, 
      sNEG		= 2110,
      sNOT		= 2120, 
      sLENGTH		= 2130, 
    sBINOP		= 2200, 
      sADD		= 2210,	// Arithmetic OPs
      sSUB		= 2211,
      sMULT		= 2212,
      sDIV		= 2213,
      sMOD		= 2214,
      sPOW		= 2215, 
      sLESS		= 2220,	// Comparators
      sEQUAL		= 2221,
      sLEEQ		= 2222,
      sNOTEQ		= 2223,
      sAND		= 2230,	// Logical OPs
      sOR		= 2231, 
      sXOR		= 2232, 
      sCONCAT		= 2240,	// Others
    sBASEFUNCT		= 2300,
      sEXP		= 2301, 
      sSIN		= 2302, // Trigonometric Functions
      sCOS		= 2303,
      sTAN		= 2304,
      sCOT		= 2305,
      sSINH		= 2306,
      sCOSH		= 2307,
      sTANH		= 2308,
      sCOTH		= 2309, 
      sABS		= 2320, // Complex -> Real
      sRE		= 2321,
      sIM		= 2322,
      sCONJ		= 2330, // Others
      sFLOOR		= 2312,
      sCEIL		= 2313,
      sSQRT		= 2314,
    sLISTFUNCT		= 2400,
      sLOG		= 2401,
      sRANDOM		= 2402,
      sMIN		= 2403,
      sMAX		= 2404,
      sGCD		= 2405,
      sLCM		= 2406,

  sSTMT			= 3000,
  sSTMTLIST		= 3010,
    sCALL		= 3020,
    sASSIGN		= 3040,
    sTRANS		= 3050,
      sFANOUT		= 3051,
      sSWAP		= 3052,
    sFOR		= 3060,
    sIF			= 3070,
    sLOOP		= 3080,
      sWHILE		= 3081,
      sUNTIL		= 3082,
    sRETURN		= 3090,
    sINPUT		= 3100,
    sPRINT		= 3110,
    sEXIT		= 3120,
    sMEASURE		= 3130,
    sRESET		= 3140,
    sINCLUDE		= 3500,
    sDUMP		= 3510, 
    sINSPECT		= 3520,
    sLOAD		= 3530,
    sSAVE		= 3540,
    sSET		= 3550,
    sSHELL		= 3560,
    
  sDEF			= 4000,
  sDEFLIST		= 4010,
    sROUTDEF		= 4100, 
      sFUNCTDEF		= 4110, 
      sPROCDEF		= 4120,
      sQUOPDEF		= 4130,
      sQUFUNDEF		= 4140,
    sARGDEF		= 4200, 
    sVARDEF		= 4210, 
    sCONSTDEF		= 4220,

  sEOF			= 5000,
  sMSG			= 5010
};

// classes in syntax.h

class sObject;
class sExpr;
class sStmt;
class sDef;
class sExprList;
class sStmtList;
class sDefList;

// classes in symbols.h

class SymTable;

// classes in quheap.h

class QuHeap;

// classes in error.h

class tError;

// classes in extern.h

class QuBaseState;
struct RoutTableEntry;

typedef int tExtRout(SymTable *,SymTable *,QuHeap *,tError *,int);

class tType {
  BaseType _type;
public:
  tType(BaseType t=tUNDEF) { _type=t; }
  tType(const tType& t) { _type=t._type; }
  int isBaseType() const { return _type>=tBOOLEAN && _type<=tQUSCR; }
  int isExpr() const { return _type>=tBOOLEAN; }
  int isDefined() const { return _type>=tVOID; }
  int isNumber() const { return _type>=tINTEGER && _type<=tCOMPLEX; }
  int isInt() const { return _type==tINTEGER; }
  int isBoolean() const { return _type==tBOOLEAN; }
  int isReal() const { return _type==tREAL; }
  int isOrdered() const { return isInt() || isReal(); }
  int isComplex() const { return _type==tCOMPLEX; }
  int isQuExpr() const { return _type>=tQUCONST && _type<=tQUSCR; }
  int isQuConst() const { return _type==tQUCONST; }
  int isQuReg() const { return _type==tQUREG; }
  int isQuVoid() const { return _type==tQUVOID; }
  int isQuScr() const { return _type==tQUSCR; }
  int isQuVar() const { return isQuExpr() && !isQuConst(); }
  int isString() const { return _type==tSTRING; }
  int isError() const { return _type==tERROR; }
  int isUndef() const { return _type==tUNDEF; }
  operator BaseType() const { return _type; }
  void operator=(const tType& t) { _type=t._type; }
  void operator=(BaseType t) { _type=t; }
  int operator==(const tType& t) const { return _type==t._type; }
  int operator==(BaseType t) const { return _type==t; }
  int operator!=(const tType& t) const { return _type!=t._type; }
  int operator!=(BaseType t) const { return _type!=t; }
  string str() const;
};

ostream& operator << (ostream& s, const tType& t);

class tValue {
  tType _type;
  union {
    boolean b;
    int i;
    tReal r;
    tComplex *pz;
    string *ps;
    quVar *pq;
  } _val;
  void delpointer();
  void assign(const tValue& v) {
    _type=v._type;
    switch(_type) {
      case tBOOLEAN:  	_val.b=v._val.b; break;
      case tINTEGER:  	_val.i=v._val.i; break;
      case tREAL:	_val.r=v._val.r; break;
      case tCOMPLEX:	_val.pz=new tComplex(*v._val.pz); break;
      case tSTRING:	_val.ps=new string(*v._val.ps); break;
      case tQUCONST:
      case tQUREG:
      case tQUVOID:
      case tQUSCR:
			_val.pq=new quVar(*v._val.pq); break;
      default:		;
    };
  }
public:
  tValue() { _type=tUNDEF; }
  tValue(BaseType t) {
     _type=t;
      switch(t) {
      case tBOOLEAN:  	_val.b=FALSE; break;
      case tINTEGER:  	_val.i=0; break;
      case tREAL:	_val.r=0.0; break;
      case tCOMPLEX:	_val.pz=new tComplex(); break;
      case tSTRING:	_val.ps=new string(); break;
      case tQUCONST:
      case tQUREG:
      case tQUVOID:
      case tQUSCR:
      			_val.pq=new quVar(); break;
      }
   }
  tValue(const tValue& v) { assign(v); }
  tValue(boolean b) { _type=tBOOLEAN; _val.b=b; }
  tValue(int i) { _type=tINTEGER; _val.i=i; }
  tValue(long i) { _type=tINTEGER; _val.i=i; }
  tValue(tReal r) { _type=tREAL; _val.r=r; }
  tValue(const tComplex& z) { _type=tCOMPLEX; _val.pz=new tComplex(z); }
  tValue(const char* s) { _type=tSTRING; _val.ps=new string(s); }
  tValue(string s) { _type=tSTRING; _val.ps=new string(s); }
  tValue(quState& q,BaseType t=tQUREG) { _type=t; _val.pq=new quVar(q); }
  tValue(quState *q,BaseType t=tQUREG) { _type=t; _val.pq=new quVar(q); }
  ~tValue() { delpointer(); }
  tType type() const { return _type; }
  int isBaseType() const { return _type.isBaseType(); }
  int isExpr() const { return _type.isExpr(); }
  int isDefined() const { return _type.isDefined(); }
  int isNumber() const { return _type.isNumber(); }
  int isBoolean() const { return _type.isBoolean(); }
  int isInt() const { return _type.isInt(); }
  int isReal() const { return _type.isReal(); }
  int isOrdered() const { return isInt() || isReal(); }
  int isComplex() const { return _type.isComplex(); }
  int isQuExpr() const { return _type.isQuExpr(); }
  int isQuConst() const { return _type.isQuConst(); }
  int isQuReg() const { return _type.isQuReg(); }
  int isQuVoid() const { return _type.isQuVoid(); }
  int isQuScr() const { return _type.isQuScr(); }
  int isQuVar() const { return isQuExpr() && !isQuConst(); }
  int isString() const { return _type.isString(); }
  int isError() const { return _type.isError(); }
  int isUndef() const { return _type.isUndef(); }
  
  int isZero() const { return (isInt() && _val.i==0) || (_type==tREAL && 
    _val.r==0.00) || (_type==tCOMPLEX && *_val.pz==tComplex(0,0)); }
  operator boolean() const { return _val.b; }
  operator int() const { return _type==tINTEGER ? _val.i : (int) _val.r; }
  operator tReal() const { return _type==tREAL ? _val.r : (tReal) _val.i; }
  operator tComplex() const { return _type==tCOMPLEX ? *_val.pz : 
			(_type==tREAL ? (tComplex) _val.r : (tComplex) _val.i);  }
  operator string() const { return *_val.ps; }
  operator quVar() const { return *_val.pq; }
  quState* qustate() const { return isQuExpr() ? _val.pq : 0; }
  operator = (const tValue& v) { delpointer(); assign(v); }
  string str() const;
};

ostream& operator << (ostream& s, const tValue& v);

#endif
